#
#   uci.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#
# timeout in line 152 set to 60 bw 8.2.17
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import GLib
import os
import subprocess
import _thread
import time

from . import engine_debug
from . import engine_output
from . import gv


class Uci:

    def __init__(self, side):
#from gi.repository import GLib
        self.engine = "jcchess"
        self.path = ""
        self.engine_running = False
        self.newgame = False
        self.running_engine = ""
        self.stop_pending = False
        self.ponder_move = None
        self.side = side
        self.engine_debug = engine_debug.get_ref()
        self.engine_output = engine_output.get_ref()

    def start_engine(self, path):

        if path is None:
            path = self.path
            # if using builtin engine return (not UCI)
            if self.engine == "jcchess":
                return

        # path is the path to the UCI engine executable
        if not os.path.isfile(path):
            print("invalid ucipath:", path)
            return False

        #
        # get engine working directory
        #from gi.repository import GLib
        orig_cwd = os.getcwd()
        if gv.verbose:
            print("current working directory is", orig_cwd)

        engine_wdir = os.path.dirname(path)
        if gv.verbose:
            print("engine working directory is", engine_wdir)

        # Attempt to start the engine as a subprocess
        if gv.verbose:
            print("starting engine with path:", path)
        path = path.strip()
        #print("engine:" + path)

        # when jcchess is started on windows with pythonw
        # a console window appears each time the engine starts
        # Use STARTUPINFO to suppress this
        if os.name == 'nt':
            si = subprocess.STARTUPINFO()
            si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            p = subprocess.Popen(
                path,bufsize = 1,   stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, cwd=engine_wdir,
                universal_newlines=True,startupinfo=si)
        else:
            p = subprocess.Popen(
                path,bufsize = 1,   stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, cwd=engine_wdir,
                universal_newlines=True)
        self.p = p
       
        #check process is running
        i = 0
        while (p.poll() is not None):
            i += 1
            if i > 40:
                print("unable to start engine process")
                return False
            time.sleep(0.25)

        if gv.verbose:
            print("pid=", p.pid)
        # start thread to read stdout
        self.op = []
        self.soutt = _thread.start_new_thread(self.read_stdout, ()) # commenting this line doesn't help

        # Tell engine to use the UCI (universal chess interface).
        self.command("uci\n")

        # wait for reply
        self.uservalues=gv.engine_manager.get_uservalues(self.engine)
        self.uci_option = []
        uci_ok = False
        i = 0
        while True:
            for l in self.op:
                l = l.strip()
                # print l
                if l.startswith("option"):
                    self.uci_option.append(self.option_parse(l))
                if l == "uciok":
                    uci_ok = True
            self.op = []
            if uci_ok:
                break
            i += 1
            if i > 40:
                print("error - uciok not returned from engine")
                return False
            time.sleep(0.25)

        # set hash value
        self.command(
            "setoption name Hash value " +
            str(gv.engine_manager.get_hash_value()) + "\n")
        
        # set pondering
        if gv.engine_manager.get_ponder():
            ponder_str = "true"
        else:
            ponder_str = "false"
        self.command("setoption name Ponder value " + ponder_str + "\n")
        
        # send setoption where we have a uservalue that differs from default
        uservalues = gv.engine_manager.get_uservalues(self.engine)
        for name, value in uservalues.items():
            self.command("setoption name " + name + " value " + value + "\n")
            
        # Ask if ready
        self.command("isready\n")

        # wait for reply
        ready_ok = False
        i = 0
        while True:
            for l in self.op:
                l = l.strip()
                # print l
                if l == "readyok":
                    ready_ok = True
            self.op = []
            if ready_ok:
                break
            i += 1
            if i > 60:
                print("error - readyok not returned from engine")
                return False
            time.sleep(0.25)

        # Tell engine we are starting new game
        self.command("ucinewgame\n")
        self.engine_running = True
        self.running_engine = self.engine
        return True

    def option_parse(self, option_line):
        name = ""
        otype = ""
        default = ""
        minimum = ""
        maximum = ""
        userval = ""
        try:
            words = option_line.split()
            w = words.pop(0)
            if w != "option":
                if gv.verbose:
                    print("invalid option line ignored:", option_line)
                return

            # get option name
            w = words.pop(0)
            if w != "name":
                if gv.verbose:
                    print("invalid option line ignored:", option_line)
                return
            # name can contain spaces
            name = ''
            w = words.pop(0)
            while w != "type" and len(words) != 0:
                name += ' ' + w
                w = words.pop(0)
            name=name.strip()

            # get option type
            if w != "type":
                if gv.verbose:
                    print("invalid option line ignored:", option_line)
                return
            otype = words.pop(0)

            uvars = []
            while True:
                w = words.pop(0)
                w2 = words.pop(0)
                if w == "default":
                    default = w2
                elif w == "min":
                    minimum = w2
                elif w == "max":
                    maximum = w2
                elif w == "var":
                    uvars.append(w2)
                elif w == "userval":
                    userval = w2
                else:
                    if gv.verbose:
                        print("error parsing option:", option_line)
        except IndexError:
            pass
        userval = self.uservalues.get(name, default)
        return([name, otype, default, minimum, maximum, uvars, userval])
                        
    def command(self, cmd):
        e = self.side + "(" + self.get_running_engine().strip() + "):"
        if gv.verbose or gv.verbose_uci:
            print("->" + e + cmd.strip())
        GObject.idle_add(self.engine_debug.add_to_log, "->" + e + cmd.strip())
        try:
            # write as string (not bytes) since universal_newlines=True
            self.p.stdin.write(cmd)
        except AttributeError:
            GObject.idle_add(
                self.engine_debug.add_to_log,
                "# engine process is not running")
        except IOError:
            GObject.idle_add(
                self.engine_debug.add_to_log,
                "# engine process is not running")

    def stop_engine(self):
        if not self.engine_running:
            return

        self.stop_pending = True
        engine_stopped = False

        try:
            if gv.verbose:
                print("stopping engine")

            self.command("quit\n")

            # allow 2 seconds for engine process to end
            i = 0
            while True:
                if self.p.poll() is not None:
                    engine_stopped = True
                    break
                i += 1
                if i > 8:
                    if gv.verbose:
                        print("engine has not terminated after quit command")
                    break
                time.sleep(0.25)

            if not engine_stopped:
                if gv.verbose:
                    print("terminating engine subprocess pid ", self.p.pid)
                # SIGTERM
                self.p.terminate()
                i = 0
                while True:
                    if self.p.poll() is not None:
                        engine_stopped = True
                        break
                    i += 1
                    if i > 8:
                        if gv.verbose:
                            print("engine has not responded to terminate " \
                                  "command")
                        break
                    time.sleep(0.25)

            if not engine_stopped:
                if gv.verbose:
                    print("killing engine subprocess pid ", self.p.pid)
                # SIGKILL
                self.p.kill()
                i = 0
                while True:
                    if self.p.poll() is not None:
                        engine_stopped = True
                        break
                    i += 1
                    if i > 16:
                        if gv.verbose:
                            print("engine has not responded to kill command")
                        print("unable to stop engine pid", self.p.pid)
                        break
                    time.sleep(0.25)
        except:
            pass

        if gv.verbose:
            print()
        if engine_stopped:
            if gv.verbose:
                print("engine stopped ok")
        self.engine_running = False
        self.stop_pending = False
        self.running_engine = ""

    def read_stdout(self): 
        while True:
            try:
                e = ("<-" + self.side + "(" +
                     self.get_running_engine().strip() + "):")
                line= ""
                # python2 line = unicode(self.p.stdout.readline(), errors ='ignore')
                    # or: 'your iso 8859-15 text'.decode('iso8859-15')
                # python3 (doesn't work) lineb = self.p.stdout.readline().encode("utf-8", "ignore")
                #print(lineb)
                #line = str(lineb)
                #print(line, "line")
                line = self.p.stdout.readline()
                
                if line == "":
                    if gv.verbose:
                        print(e + "eof reached")
                    if gv.verbose:
                        print(e + "stderr:", self.p.stderr.read())
                    break
                #line = line[2:-3]
                #print(line)
                line = line.strip()
                
                if gv.verbose or gv.verbose_uci:
                    print(e + line)
                GObject.idle_add(self.engine_debug.add_to_log, e+line)
                if line.startswith("info"):
                    GObject.idle_add(
                        self.engine_output.add_to_log, self.side,
                        self.get_running_engine().strip(), line)
                self.op.append(line)
            except Exception as e:
                # line = e + "error"
                print("subprocess error in uci.py read_stdout:", e, "at:", line)
                

    def check_running(self):
        # check if engine has changed since last use
        if self.engine != self.running_engine:
            if self.engine_running:
                self.stop_engine()
                self.start_engine(None)
                return

        if not self.engine_running:
            self.start_engine(None)
        else:
            if self.p.poll() is not None:
                print("warning engine has stopped running - attempting " \
                      "to restart")
                self.start_engine(None)

    def set_newgame(self):
        self.newgame = True

    # Ask engine to make move
    def cmove(self, movelist, side_to_move):
        self.check_running()

        if self.newgame:
            self.command("ucinewgame\n")
            self.newgame = False

        startpos = gv.jcchess.get_startpos()

        # if not startpos must be fen
        if startpos != "startpos":
            startpos = "fen " + startpos

        ml = ""
        for move in movelist:
            ml = ml + move + " "
        if ml == "":
            b = "position " + startpos + "\n"
        else:
            b = "position " + startpos + " moves " + ml.strip() + "\n"

        # b = "position sfen lnsgkgsnl/1r5b1/ppppppppp/9/9/9/PPPPPPPPP/" \
        #     "1B5R1/LNSGKGSNL b - 1 moves " + ml.strip() +  "\n"
        # b = "position sfen 8l/1l+R2P3/p2pBG1pp/kps1p4/Nn1P2G2/P1P1P2PP/" \
        #     "1PS6/1KSG3+r1/LN2+p3L w Sbgn3p 124 moves " + ml.strip() +  "\n"

        # Send the board position to the engine
        self.command(b)

        # times in milliseconds
        # btime = time_left[0]
        # wtime = time_left[1]
        # byoyomi = time_left[2]

        # clear the engine output window ready for next move
        GObject.idle_add(
            self.engine_output.clear, self.side,
            self.get_running_engine().strip())

        # print "calling time control module from uci module to get go command"
        gocmnd = gv.tc.get_go_command(side_to_move)
        self.gocmnd = gocmnd        # save for possible ponder
        # print "go command:", gocmnd

        # start the clock
        # print "starting clock from uci.py"
        gv.tc.start_clock(side_to_move)

        # send the engine the command to do the move
        self.command(gocmnd + "\n")

        # self.command(
        #   "go btime " + str(btime) +" wtime " + str(wtime) + " byoyomi " +
        #   str(byoyomi) + "\n")

        # Wait for move from engine
        i = 0
        bestmove = False
        while True:

            time.sleep(0.5)

            # i += 1
            # print "in cmove i=",i

            # if stop command sent while engine was thinking then return
            if not self.engine_running or self.stop_pending:
                return None, None

            for l in self.op:
                l = l.strip()
                if l.startswith("bestmove"):
                    bestmove = l[9:].strip()
                    if gv.verbose:
                        print("bestmove is ", bestmove)

                    # get ponder move if present
                    self.ponder_move = None
                    s = bestmove.find("ponder")
                    if s != -1:
                        self.ponder_move = bestmove[s + 7:].strip()
                        GObject.idle_add(
                            self.engine_output.set_ponder_move,
                            # set ponder move in engine output window
                            self.ponder_move, self.side)

                    # get bestmove
                    s = bestmove.find(" ")
                    if s != -1:
                        bestmove = bestmove[:s]
                    self.op = []

                    # update time for last move
                    GObject.idle_add(gv.tc.update_clock)
                    GObject.idle_add(gv.gui.set_side_to_move, side_to_move)
                    # Allow above update_clock to complete
                    # This must complete before we do start_clock for 
                    # human in jcchess.py  
                    time.sleep(0.1)

                    return bestmove, self.ponder_move
            self.op = []

    def stop_ponder(self):
        # return if not pondering
        # if self.ponder_move is None:
        #    return
        # stop pondering
        self.command("stop\n")
        # Wait for move from engine
        i = 0
        bestmove = False
        while True:

            time.sleep(0.5)

            # i += 1
            # print "in stop ponder i=",i

            # if stop command sent while engine was thinking then return
            if not self.engine_running or self.stop_pending:
                return None, None

            for l in self.op:
                l = l.strip()
                if l.startswith("bestmove"):
                    bestmove = l[9:].strip()
                    if gv.verbose:
                        print("ponder bestmove is ", bestmove)

                    # get ponder move if present
                    ponder_move = None
                    s = bestmove.find("ponder")
                    if s != -1:
                        ponder_move = bestmove[s + 7:].strip()

                    # get bestmove
                    s = bestmove.find(" ")
                    if s != -1:
                        bestmove = bestmove[:s]
                    self.op = []

                    return bestmove, ponder_move
            self.op = []

    def send_ponderhit(self, side_to_move):

        # start the clock
        gv.tc.start_clock(side_to_move)

        self.command("ponderhit\n")
        # Wait for move from engine
        i = 0
        bestmove = False
        while True:

            time.sleep(0.5)

            # i += 1
            # print "in send ponder i,side=",i,self.side

            # if stop command sent while engine was thinking then return
            if not self.engine_running or self.stop_pending:
                return None, None

            for l in self.op:
                l = l.strip()
                if l.startswith("bestmove"):
                    bestmove = l[9:].strip()
                    if gv.verbose:
                        print("bestmove is ", bestmove)

                    # get ponder move if present
                    self.ponder_move = None
                    s = bestmove.find("ponder")
                    if s != -1:
                        self.ponder_move = bestmove[s + 7:].strip()
                        GObject.idle_add(
                            self.engine_output.set_ponder_move,
                            # set ponder move in engine output window
                            self.ponder_move, self.side)

                    # get bestmove
                    s = bestmove.find(" ")
                    if s != -1:
                        bestmove = bestmove[:s]
                    self.op = []

                    # update time for last move
                    # print "updating clock from uci.py"
                    GLib.idle_add(gv.tc.update_clock)
                    GLib.idle_add(gv.gui.set_side_to_move, side_to_move)

                    return bestmove, self.ponder_move
            self.op = []

    def start_ponder(self, pondermove, movelist, cmove):

        startpos = gv.jcchess.get_startpos()

        # if not startpos must be sfen
        if startpos != "startpos":
            startpos = "fen " + startpos

        ml = ""
        for move in movelist:
            ml = ml + move + " "
        # if ml == "":
        #    print "error empty movelist in ponder in uci.py"
        #    return

        ml = ml.strip()
        ml = ml + " " + cmove + " " + pondermove
        ml = ml.strip()

        # create the position string with the ponder move added
        b = "position " + startpos + " moves " + ml + "\n"

        # Send the board position to the engine
        self.command(b)

        pondercmd = "go ponder" + self.gocmnd[2:]
        self.command(pondercmd + "\n")

        # clear the engine output window ready for next move
        GObject.idle_add(
            self.engine_output.clear, self.side,
            self.get_running_engine().strip())

        return

    def get_options(self):
        return self.uci_option

    def set_options(self, options):
        self.uci_option = options

    def set_option(self, option):
        option = option + "\n"
        self.command(option)

    # used when adding new engines in engine_manager
    def test_engine(self, path):
        msg = ""
        name = ""

        # path is the path to the UCI engine executable
        if not os.path.isfile(path):
            msg = "invalid path " + path
            return msg, name

        running = self.engine_running
        if running:
            self.stop_engine()

        # Attempt to start the engine as a subprocess
        engine_wdir = os.path.dirname(path)
        try:
            if os.name == 'nt':
                si = subprocess.STARTUPINFO()
                si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                p = subprocess.Popen(
                    path,bufsize = 1,   stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE, cwd=engine_wdir,
                    universal_newlines=True,startupinfo=si)
            else:
                p = subprocess.Popen(
                    path,bufsize = 1,   stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE, cwd=engine_wdir,
                    universal_newlines=True)
        except OSError as oe:
            msg = "error starting engine: " + "OSError" + str(oe)
            return msg, name
        self.p = p

        # check process is running
        i = 0
        while (p.poll() is not None):
            i += 1
            if i > 40:
                msg = "not a valid UCI engine"
                return msg, name
            time.sleep(0.25)

        # start thread to read stdout commented out (double?)
        self.op = []
        self.soutt = _thread.start_new_thread(self.read_stdout, ())

        # Tell engine to use the UCI (universal chess interface).
        self.command("uci\n")

        # wait for reply
        self.uservalues=gv.engine_manager.get_uservalues(self.engine)
        self.uci_option = []
        uci_ok = False
        i = 0
        while True:
            for l in self.op:
                l = l.strip()
                if l.startswith("id "):
                    w = l.split()
                    if w[1] == "name":
                        w.pop(0)
                        w.pop(0)
                        for j in w:
                            name = name + j + " "
                        name = name.strip()
                elif l.startswith("option"):        
                    self.uci_option.append(self.option_parse(l))
                elif l == "uciok":
                    uci_ok = True
            self.op = []
            if uci_ok:
                break
            i += 1
            if i > 40:
                msg = "not a valid UCI engine"
                return msg, name
            time.sleep(0.25)

        try:
            self.command("quit\n")
            self.p.terminate()
        except:
            pass
        return msg, name

    def set_engine(self, ename, path):
        self.engine = ename
        if path is None:
            self.path = gv.engine_manager.get_path(ename)
        else:
            self.path = path

    def set_path(self, epath):
        self.path = epath

    def get_engine(self):
        return self.engine

    def get_running_engine(self):
        if self.running_engine == "":
            return self.engine
        else:
            return self.running_engine

    def UCI_options(self):
        self.check_running()
        options = self.get_options()
        # option
        # [name, otype, default, minimum, maximum, uvars, userval]
        wdgts = []
        opt_i = -1
        for option in options:
            opt_i += 1
            name = option[0]
            otype = option[1]
            default = option[2]
            minimum = option[3]
            maximum = option[4]
            uvars = option[5]
            userval = option[6]
            # if in common engine settings then skip
            if name in ("Hash", "Ponder"):
                continue
            wdgts.append((
                opt_i, name, otype, default, minimum, maximum, uvars, userval))

        diagtitle = self.get_engine()
        dialog = Gtk.Dialog(
            diagtitle, gv.gui.get_window(), 0,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OK, Gtk.ResponseType.OK))
        sw = Gtk.ScrolledWindow.new(None, None)
        dialog.vbox.pack_start(sw, True, True, 0)

        vb=Gtk.Box.new(Gtk.Orientation.VERTICAL, 0)
        sw.add(vb)

        sw.show()
        vb.show()

        wlist = []
        for w in wdgts:
            opt_i, name, otype, default, minimum, maximum, uvars, userval = w
            #print(name)
            if otype == "spin":
                if minimum == "":
                    minimum = 0
                if maximum == "":
                    maximum = 10
                if default == "":
                    default = minimum
                if userval != "":
                    default = userval
                adj = Gtk.Adjustment(
                    value=float(default), lower=float(minimum), upper=float(maximum),
                    step_increment=1, page_increment=5, page_size=0)
                spinner = Gtk.SpinButton.new(adj, 1.0, 0)
                # spinner.set_width_chars(14)
                al = Gtk.Alignment.new(
                    xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
                al.add(spinner)

                lbl = Gtk.Label(label=name + ":")

                hb = Gtk.HBox(False, 0)
                hb.pack_start(lbl, False, False, 0)
                hb.pack_start(al, True, True, 10)

                vb.pack_start(hb, False, True, 0)

                v = (opt_i, adj, name, otype)
                wlist.append(v)

                lbl.show()
                spinner.show()
                al.show()
                hb.show()
            elif otype == "string":
                ent = Gtk.Entry()
                if userval != "":
                    default = userval
                ent.set_text(default)

                al = Gtk.Alignment.new(
                    xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
                al.add(ent)

                lbl = Gtk.Label(label=name + ":")

                hb = Gtk.HBox(False, 0)
                hb.pack_start(lbl, False, False, 0)
                hb.pack_start(al, True, True, 10)

                vb.pack_start(hb, False, True, 0)

                v = (opt_i, ent, name, otype)
                wlist.append(v)

                lbl.show()
                ent.show()
                al.show()
                hb.show()
            elif otype == "check":
                cb = Gtk.CheckButton(label=None)
                if userval != "":
                    default = userval
                if default == "true":
                    cb.set_active(True)
                else:
                    cb.set_active(False)
                al = Gtk.Alignment.new(
                    xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
                al.add(cb)

                lbl = Gtk.Label(label=name + ":")
                hb = Gtk.HBox(False, 0)
                hb.pack_start(lbl, False, False, 0)
                hb.pack_start(al, True, True, 10)

                vb.pack_start(hb, False, True, 0)

                v = (opt_i, cb, name, otype)
                wlist.append(v)

                lbl.show()
                cb.show()
                al.show()
                hb.show()
            elif otype == "combo":
                if userval != "":
                    default = userval
                combobox = Gtk.ComboBoxText()
                i = 0
                for v in uvars:
                    combobox.append_text(v)
                    if v == default:
                        combobox.set_active(i)
                    i += 1

                al = Gtk.Alignment.new(
                    xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
                al.add(combobox)
                al.show()
                combobox.show()

                lbl = Gtk.Label(label=name + ":")
                hb = Gtk.HBox(False, 0)
                hb.pack_start(lbl, False, False, 0)
                hb.pack_start(al, True, True, 10)

                vb.pack_start(hb, False, True, 0)

                v = (opt_i, combobox, name, otype)
                wlist.append(v)

                lbl.show()
                hb.show()
            elif otype == "button":
                b = Gtk.Button(label=name)
                al = Gtk.Alignment.new(
                    xalign=1.0, yalign=0.0, xscale=0.0, yscale=0.0)
                al.add(b)

                b.connect(
                    "clicked",
                    lambda widget: self.command(
                        "setoption name " +
                        widget.get_label() + "\n"))

                lbl = Gtk.Label(label=name + ":")
                hb = Gtk.HBox(False, 0)
                hb.pack_start(lbl, False, False, 0)
                hb.pack_start(al, True, True, 10)
                vb.pack_start(hb, False, True, 0)
                lbl.show()
                b.show()
                al.show()
                hb.show()
            else:
                if gv.verbose:
                    print("type ignored - ", otype)
                  
        # set size of scrollwindow to size of child            
        minsize, naturalsize = vb.get_preferred_size()
        sw.set_size_request(minsize.width+20, min(minsize.height, 500))

        dialog.set_default_response(Gtk.ResponseType.OK)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            for w in wlist:
                opt_i, widge, name, otype = w
                if otype == "spin":
                    av = int(widge.get_value())
                elif otype == "string":
                    av = widge.get_text()
                elif otype == "check":
                    if widge.get_active():
                        av = "true"
                    else:
                        av = "false"
                elif otype == "combo":
                    av = widge.get_active_text()
                else:
                    if gv.verbose:
                        print("unknown type", otype)
                # setoption name <id> [value <x>]
                # uci.set_option(
                #   "option name LimitDepth type spin default 10 min 4 max 10")
                a = "setoption name " + name + " value " + str(av)
                self.set_option(a)
                # update user value
                options[opt_i][6] = str(av)
            self.set_options(options)
            gv.engine_manager.set_uservalues(self.engine, options)
        dialog.destroy()
        self.stop_engine()



#
#   gui.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import GdkPixbuf
from gi.repository import Pango
import os
import cairo
from datetime import date

from . import engine_debug
from . import engine_output
from . import move_list
from . import drag_and_drop
from . import load_save
from . import utils
from . import gamelist
from . import comments
import chess
from .constants import WHITE, BLACK, NAME, VERSION, TARGET_TYPE_TEXT
from . import gv


class Gui:

    gui_ref = None

    def __init__(self):
        if Gui.gui_ref is not None:
            raise RuntimeError("Attempt to create second gui instance")
        Gui.gui_ref = self

    def build_gui(self):

        self.gobactive = False        
        glade_dir = gv.jcchess.get_glade_dir() 
        self.promotion_glade_file = os.path.join(glade_dir, "promotion.glade")
        self.promotion_piece = 'q'
        self.engine_debug = engine_debug.get_ref()
        self.engine_output = engine_output.get_ref()
        self.move_list = move_list.get_ref()
        self.gamelist = gamelist.get_ref()
        self.drag_and_drop = drag_and_drop.get_ref()
        self.enable_dnd = True
        self.show_coords = True
        self.highlight_moves = True
        self.lastdir = os.path.expanduser("~") # Filehandling
        # Create Main Window
        glade_dir = gv.jcchess.get_glade_dir()
        self.glade_file = os.path.join(glade_dir, "main_window.glade")
        self.glade_file_preferences = os.path.join(
            glade_dir, "preferences.glade")
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file)
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("main_window")
        screen = self.window.get_screen()
        self.screen_width = screen.get_width()
        self.screen_height = screen.get_height()
        self.set_window_size()

     
        self.window.set_title(NAME + " " + VERSION)

        # 1 eventbox per board square
        self.eb = [[Gtk.EventBox() for x in range(8)] for x in range(8)]

        # Set a handler for delete_event that immediately exits GTK.
        self.window.connect("delete_event", gv.jcchess.delete_event)

        main_vbox = self.builder.get_object("main_vbox")
        main_vbox.show()

        # menu
        # Create a UIManager instance
        uimanager = Gtk.UIManager()

        # Add the accelerator group to the toplevel window
        accelgroup = uimanager.get_accel_group()
        self.window.add_accel_group(accelgroup)

        # main action group
        actiongroup = Gtk.ActionGroup("UIManagerAG")
        self.actiongroup = actiongroup

        # action group for toggleactions
        ta_action_group = Gtk.ActionGroup("AGPromoteMode")
        ta_action_group.add_toggle_actions([(
            "promotemode", None, _("_Ask Before Promoting"), None, None,
            self.promote_mode)])
             
        self.ta_action_group = ta_action_group

        # Create actions
        actiongroup.add_actions([
            # NewGame and the handicap names are used in jcchess.py NewGame
            # routine. don't change them.
            ("NewGame", Gtk.STOCK_NEW, _("_New Game"), None, _("New Game"),
             gv.jcchess.new_game_cb),
            ("LanceHandicap", None, _("_Lance"), None,
             _("Lance Handicap"), gv.jcchess.new_game_cb),
            ("BishopHandicap", None, _("_Bishop"), None,
             _("Bishop Handicap"), gv.jcchess.new_game_cb),
            ("RookHandicap", None, _("_Rook"), None,
             _("Rook Handicap"), gv.jcchess.new_game_cb),
            ("RookandLanceHandicap", None, _("_Rook and Lance"), None,
             _("Rook and Lance Handicap"), gv.jcchess.new_game_cb),
            ("TwoPieceHandicap", None, _("_2 Pieces"), None,
             _("Two Piece Handicap"), gv.jcchess.new_game_cb),
            ("FourPieceHandicap", None, _("_4 Pieces"), None,
             _("Four Piece Handicap"), gv.jcchess.new_game_cb),
            ("SixPieceHandicap", None, _("_6 Pieces"), None,
             _("Six Piece Handicap"), gv.jcchess.new_game_cb),
            ("EightPieceHandicap", None, _("_8 Pieces"), None,
             _("Eight Piece Handicap"), gv.jcchess.new_game_cb),
            ("TenPieceHandicap", None, _("_10 Pieces"), None,
             _("Ten Piece Handicap"), gv.jcchess.new_game_cb),
            ("ThreePawnHandicap", None, _("_3 Pawns"), None,
            _("Three Pawn Handicap"), gv.jcchess.new_game_cb),
            ("NewHandicapGame", None, _("_New Handicap Game")),
            #
            ("Quit", Gtk.STOCK_QUIT, _("_Quit"), None,
             _("Quit the Program"), gv.jcchess.quit_game),
            ("LoadGame", Gtk.STOCK_OPEN, _("_Load Game"), None,
             _("Load Game"), gv.load_save.load_game),
            ("SaveGame", Gtk.STOCK_SAVE, _("_Save Game"), None,
             _("Save Game"), gv.load_save.save_game),
            ("File", None, _("_File")),
            ("Edit", None, _("_Edit")),
            ("Undo", Gtk.STOCK_UNDO, _("_Undo Move"), "<Control>U",
             _("Undo Move"), gv.jcchess.undo_single_move),
            ("Redo", Gtk.STOCK_REDO, _("_Redo Move"), "<Control>R",
             _("Redo Move"), gv.jcchess.redo_single_move),
            ("MoveNow", None, _("_Move Now"), "<Control>M",
             _("Move Now"), gv.jcchess.move_now),
            ("SetBoardColours", None, _("_Set Board Colours"), None,
             _("Set Board Colours"), gv.set_board_colours.show_dialog),
            ("SetPieces", None, _("_Set Pieces"), None,
             _("Set Pieces"), gv.pieces.show_pieces_dialog),
            ("TimeControl", None, _("_Time Control"), None,
             _("Time Control"), gv.tc.time_control),
            # ConfigureEngine1 - this name is used in engine_manager.
            # don't change it.
            ("ConfigureEngine1", None, _("_Configure Engine 1 (White)"), None,
             _("Configure Engine 1"), gv.engine_manager.configure_engine),
            ("ConfigureEngine2", None, _("_Configure Engine 2 (Black)"), None,
             _("Configure Engine 2"), gv.engine_manager.configure_engine),
            ("Players", None, _("_Players"), None,
             _("Players"), gv.jcchess.set_players),
            ("Engines", None, _("_Engines"), None,
             _("Engines"), gv.engine_manager.engines),
            ("CommonEngineSettings", None, _("_Common Engine Settings"), None,
             _("Common Engine Settings"), gv.engine_manager.common_settings),
            ("MoveList", None, _("_Move List"), None,
             _("Move List"), self.move_list.show_movelist_window),
            ("GameList", None, _("_Game List"), None,
             _("Game List"), self.gamelist.show_gamelist_window_cb),
            ("EngineOutput", None, _("_Engine Output"), None,
             _("Engine Output"), self.engine_output.show_engine_output_window),
            ("EngineDebug", None, _("_Engine Debug"), None,
             _("Engine Debug"), self.engine_debug.show_debug_window),
            ("Options", None, _("_Options")),
            ("View", None, _("_View")),
            ("About", Gtk.STOCK_ABOUT, _("_About"), None,
             _("Show About Box"), self.about_box),
            ("Help", None, _("_Help")),
            ("CopyPosition", None, _("_Copy Position"), None,
             _("Copy Position"), utils.copy_FEN_to_clipboard),
            ("PastePosition", None, _("_Paste Position"), None,
             _("Paste Position"), utils.paste_clipboard_to_FEN),
            ("CopyGame", None, _("_Copy Game"), None,
             _("Copy Game"), utils.copy_game_to_clipboard),
            ("PasteGame", None, _("_Paste Game"), None,
             _("Paste Game"), utils.paste_game_from_clipboard),
            ("EditPosition", None, _("_Edit Position"), None,
             _("Edit Position"), self.enable_edit_mode),
            ("Preferences", None, _("_Preferences"), None,
             _("Preferences"), self.preferences),
            ]
        )

        # Add the actiongroups to the uimanager
        uimanager.insert_action_group(actiongroup, 0)
        uimanager.insert_action_group(ta_action_group, 1)

        #ui = """<ui>
        #<menubar name="MenuBar">
        #    <menu action="File">
        #        <menuitem action="NewGame"/>
        #        <menu action="NewHandicapGame">
        #            <menuitem action="LanceHandicap"/>
        #            <menuitem action="BishopHandicap"/>
        #            <menuitem action="RookHandicap"/>
        #            <menuitem action="RookandLanceHandicap"/>
        #            <menuitem action="TwoPieceHandicap"/>
        #            <menuitem action="FourPieceHandicap"/>
        #            <menuitem action="SixPieceHandicap"/>
        #            <menuitem action="EightPieceHandicap"/>
        #            <menuitem action="TenPieceHandicap"/>
        #            <menuitem action="ThreePawnHandicap"/>
        #        </menu>
        #        <separator/>
        #        <menuitem action="LoadGame"/>
        #        <menuitem action="SaveGame"/>
        #        <separator/>
        #        <menuitem action="Quit"/>
        #    </menu>
        #    <menu action="Edit">
        #        <menuitem action="CopyPosition"/>
        #        <menuitem action="PastePosition"/>
        #        <separator/>
        #        <menuitem action="CopyGame"/>
        #        <menuitem action="PasteGame"/>
        #        <separator/>
        #        <menuitem action="EditPosition"/>
        #        <separator/>
        #        <menuitem action="Preferences"/>
        #    </menu>
        #    <menu action="Options">
        #        <menuitem action="Undo"/>
        #        <menuitem action="Redo"/>
        #        <menuitem action="MoveNow"/>
        #        <separator/>
        #        <menuitem action="promotemode"/>
        #        <menuitem action="SetBoardColours"/>
        #        <menuitem action="SetPieces"/>
        #        <separator/>
        #        <menuitem action="TimeControl"/>
        #        <menuitem action="ConfigureEngine1"/>
        #        <menuitem action="ConfigureEngine2"/>
        #        <menuitem action="Players"/>
        #        <menuitem action="Engines"/>
        #        <menuitem action="CommonEngineSettings"/>
        #    </menu>
        #    <menu action="View">
        #        <menuitem action="MoveList"/>
        #        <menuitem action="GameList"/>
        #        <menuitem action="EngineOutput"/>
        #        <menuitem action="EngineDebug"/>
        #    </menu>
        #    <menu action="Help">
        #        <menuitem action="About"/>
        #    </menu>
        #</menubar>
        #</ui>"""
        
        ui = """<ui>
        <menubar name="MenuBar">
            <menu action="File">
                <menuitem action="NewGame"/>
                <separator/>
                <menuitem action="LoadGame"/>
                <menuitem action="SaveGame"/>
                <separator/>
                <menuitem action="Quit"/>
            </menu>
            <menu action="Edit">
                <menuitem action="CopyPosition"/>
                <menuitem action="PastePosition"/>
                <separator/>
                <menuitem action="CopyGame"/>
                <menuitem action="PasteGame"/>
                <separator/>
                <menuitem action="EditPosition"/>
                <separator/>
                <menuitem action="Preferences"/>
            </menu>
            <menu action="Options">
                <menuitem action="Undo"/>
                <menuitem action="Redo"/>
                <menuitem action="MoveNow"/>
                <separator/>
                <menuitem action="SetBoardColours"/>
                <menuitem action="SetPieces"/>
                <separator/>
                <menuitem action="TimeControl"/>
                <menuitem action="ConfigureEngine1"/>
                <menuitem action="ConfigureEngine2"/>
                <menuitem action="Players"/>
                <menuitem action="Engines"/>
                <menuitem action="CommonEngineSettings"/>
            </menu>
            <menu action="View">
                <menuitem action="MoveList"/>
                <menuitem action="GameList"/>
                <menuitem action="EngineOutput"/>
                <menuitem action="EngineDebug"/>
            </menu>
            <menu action="Help">
                <menuitem action="About"/>
            </menu>
        </menubar>
        </ui>"""

        # Add a UI description
        uimanager.add_ui_from_string(ui)

        # Use an event box and set its background colour.
        # This was needed on Ubuntu 12.04 to set the toolbar bg colour.
        # Otherwise ubuntu uses the window bg colour which is not
        # correct.
        # don't need this on Fedora though.
        eb_1 = self.builder.get_object("eb_1")
        # eb_1 = Gtk.EventBox()
        vbox2 = Gtk.VBox(False, 0)
        # main_vbox.pack_start(vbox2, False)
        # main_vbox.pack_start(eb_1, False)
        eb_1.add(vbox2)
        eb_1.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse("#EDECEB"))##EDECEB
        # Create a MenuBar
        menubar = uimanager.get_widget("/MenuBar")
        vbox2.pack_start(menubar, False, True, 0)

        # create a toolbar
        toolbar = Gtk.Toolbar()
        vbox2.pack_start(toolbar, False, True, 0)

        # populate toolbar
        toolitem = Gtk.ToolItem()

        # 2 rows, 4 columns, not homogeneous
        tb = Gtk.Table(5, 8, False)

        # Gtk.ShadowType.NONE, SHADOW_IN, SHADOW_OUT, SHADOW_ETCHED_IN,
        # SHADOW_ETCHED_OUT

        self.side_to_move = [
            Gtk.Arrow(Gtk.ArrowType.RIGHT, Gtk.ShadowType.ETCHED_IN),
            Gtk.Arrow(Gtk.ArrowType.RIGHT, Gtk.ShadowType.ETCHED_IN)]
        self.side_to_move[BLACK].set_alignment(0.5, 0.4)
        self.side_to_move[WHITE].set_alignment(0.5, 0.4)

        tb.attach(self.side_to_move[WHITE], 0, 1, 1,2)
        tb.attach(self.side_to_move[BLACK], 0, 1, 2,3)
        #self.side_to_move.set_tooltip_text(_("Header from loaded file"))

        ue1 =Gtk.Label(label=_("Active:"))
        lw = Gtk.Label(label=_("White ") + ": ")
        lb = Gtk.Label(label=_("Black ") + ": ")
        ue1.set_alignment(0, 0.5)
        lw.set_alignment(0, 0.5)
        lb.set_alignment(0, 0.5)
        tb.attach(ue1, 1, 2, 0, 1)
        tb.attach(lw, 1, 2, 1, 2)
        tb.attach(lb, 1, 2, 2, 3)

        self.engines_lblw = Gtk.Label(label=" ") # This blank is mandatory, otherwise a fileheader will be written, reason not clear b.wille
        self.engines_lblw.set_use_markup(True)
        self.engines_lblw.set_alignment(0, 0.5)

        self.engines_lblb = Gtk.Label(label=_(" "))
        self.engines_lblb.set_use_markup(True)
        self.engines_lblb.set_alignment(0, 0.5)

        tb.attach(self.engines_lblw, 2, 3, 1, 2)
        tb.attach(self.engines_lblb, 2, 3, 2, 3)
        
        #Engines
        self.engines_lblw = Gtk.Label(label="jcchess")
        self.engines_lblw.set_use_markup(True)
        self.engines_lblw.set_alignment(0, 0.5)

        self.engines_lblb = Gtk.Label(label=_("human"))
        self.engines_lblb.set_use_markup(True)
        self.engines_lblb.set_alignment(0, 0.5)

        tb.attach(self.engines_lblw, 2, 3, 1, 2)
        tb.attach(self.engines_lblb, 2, 3, 2, 3)
        # time control
        self.tc_lbl = [
            Gtk.Label(label="00:45:00 00/10"),
            Gtk.Label(label="00:45:00 00/10")]
        self.tc_lbl[BLACK].set_alignment(0, 0.5)
        self.tc_lbl[WHITE].set_alignment(0, 0.5)

        tb.attach(self.tc_lbl[WHITE], 3, 4, 1, 2)
        tb.attach(self.tc_lbl[BLACK], 3, 4, 2, 3)
        tb.set_tooltip_text(_("Side to move(>), engines and time"))
        toolitem.add(tb)
        toolbar.insert(toolitem, -1)          
        if  gv.show_header == True:
            # add a vertical separator
            hb = Gtk.HBox(False, 0)
            toolitem = Gtk.ToolItem()
            vsep = Gtk.VSeparator()
            hb.pack_start(vsep, True, True, 10)
            toolitem.add(hb)
            toolbar.insert(toolitem, -1)               

              

        # Header
        tb = Gtk.Table(5, 8, False)
        if  gv.show_header == True:                 
    
            self.header_lue = Gtk.Label(label=_("     From file: "))
            self.header_lsente = Gtk.Label(label=_("     Sente(Blk)") + ": ")
            self.header_lgote = Gtk.Label(label= _("     Gote (Whi)") + ": ")
            self.header_levent = Gtk.Label(label=_("     Event") + ": ")
            self.header_ldate = Gtk.Label(label=_("     Date") + ": ")
            self.header_lue.set_alignment(0, 0.5)
            self.header_lsente.set_alignment(0, 0.5)
            self.header_lgote.set_alignment(0, 0.5)
            self.header_levent.set_alignment(0, 0.5)
            self.header_ldate.set_alignment(0, 0.5)
            tb.attach(self.header_lue, 5, 6, 0, 1)
            tb.attach(self.header_levent, 5, 6, 1, 2)
            tb.attach(self.header_ldate, 5, 6, 2, 3)
            tb.attach(self.header_lsente, 5, 6, 4,5)
            tb.attach(self.header_lgote, 5,  6, 3, 4)
            self.header_lblsente = Gtk.Label(label=_(" ") )
            self.header_lblgote = Gtk.Label(label=_(" ")  )
            self.header_lblevent = Gtk.Label(label=_("None"))
            self.header_lbldate = Gtk.Label(label=str(date.today().day)+ _(".")+ str(date.today().month) + _(".") +str(date.today().year))
            self.header_lblsente.set_alignment(0, 0.5)
            self.header_lblgote.set_alignment(0, 0.5)
            self.header_lblevent.set_alignment(0, 0.5)
            self.header_lbldate.set_alignment(0, 0.5)
            tb.attach(self.header_lblevent, 7, 8, 1, 2)
            tb.attach(self.header_lbldate, 7, 8, 2, 3)
            tb.attach(self.header_lblsente, 7,  8, 4, 5)
            tb.attach(self.header_lblgote, 7, 8, 3, 4)
            tb.set_tooltip_text(_("Header from loaded file"))
            toolitem = Gtk.ToolItem()
            toolitem.add(tb)
            toolbar.insert(toolitem, -1)


        # add a vertical separator
        hb = Gtk.HBox(False, 0)
        toolitem = Gtk.ToolItem()
        vsep = Gtk.VSeparator()
        hb.pack_start(vsep, True, True, 10)
        toolitem.add(hb)
        toolbar.insert(toolitem, -1)

        # stop/go buttons
        hb = Gtk.HBox(False, 0)
        self.gobutton = Gtk.ToolButton(Gtk.STOCK_YES)
        self.gobutton.connect("clicked", gv.jcchess.go_clicked)
        self.stopbutton = Gtk.ToolButton(Gtk.STOCK_NO)
        self.stopbutton.connect("clicked", gv.jcchess.stop_clicked)
        self.gobutton.set_tooltip_text(_("go"))
        self.stopbutton.set_tooltip_text(_("stop"))
        hb.pack_start(self.stopbutton, False, True, 0)
        hb.pack_start(self.gobutton, False, True, 0)
        
        toolitem = Gtk.ToolItem()
        toolitem.add(hb)
        toolbar.insert(toolitem, -1)
        
        # add a vertical separator
        hb = Gtk.HBox(False, 0)
        toolitem = Gtk.ToolItem()
        vsep = Gtk.VSeparator()
        hb.pack_start(vsep, True, True, 10)
        toolitem.add(hb)
        toolbar.insert(toolitem, -1)

        # game review buttons
        
        hb = Gtk.HBox(False, 0)
        self.go_first = Gtk.ToolButton(Gtk.STOCK_GOTO_FIRST)
        self.go_first.connect("clicked", gv.jcchess.undo_all)

        self.go_back = Gtk.ToolButton(Gtk.STOCK_GO_BACK)
        self.go_back.connect("clicked", gv.jcchess.undo_single_move)

        self.go_forward = Gtk.ToolButton(Gtk.STOCK_GO_FORWARD)
        self.go_forward.connect("clicked", gv.jcchess.redo_single_move)

        self.go_last = Gtk.ToolButton(Gtk.STOCK_GOTO_LAST)
        self.go_last.connect("clicked", gv.jcchess.redo_all)
        hb.set_tooltip_text(_("Moves"))
        hb.pack_start(self.go_first, False, True, 0)
        hb.pack_start(self.go_back, False, True, 0)
        hb.pack_start(self.go_forward, False, True, 0)
        hb.pack_start(self.go_last, False, True, 0)

        toolitem = Gtk.ToolItem()
        toolitem.add(hb)
        toolbar.insert(toolitem, -1)
                
        # add a vertical separator
        hb = Gtk.HBox(False, 0)
        toolitem = Gtk.ToolItem()
        vsep = Gtk.VSeparator()
        hb.pack_start(vsep, True, True, 10)
        toolitem.add(hb)
        toolbar.insert(toolitem, -1)

        main_grid = self.builder.get_object("grid1")
        main_grid.set_row_homogeneous(True)
        main_grid.set_column_homogeneous(True)
       #Insert comment-editing buttons
        if gv.show_moves == True:
            hb = Gtk.HBox(False, 0)
            self.cedit = Gtk.ToolButton(Gtk.STOCK_EDIT)
            self.cedit.connect("clicked", self.set_cedit)
            hb.set_tooltip_text(_("Comments"))
            self.csave = Gtk.ToolButton(Gtk.STOCK_SAVE)
            self.csave.connect("clicked", self.set_csave)
        
            self.ccancel = Gtk.ToolButton(Gtk.STOCK_CANCEL)
            self.ccancel.connect("clicked", self.set_ccancel)
            #self.gobutton.set_tooltip_text(_("go"))
            #self.gobutton.set_tooltip_text(_("go"))
            hb.pack_start(self.cedit, False, True, 0)
            hb.pack_start(self.ccancel, False, True, 0)
            hb.pack_start(self.csave, False, True, 0)
            #hb.gtk_widget_hide(self.ccancel)
            #hb.gtk_widget_hide(self.csafe)
               
            toolitem = Gtk.ToolItem()
            toolitem.add(hb)
            toolbar.insert(toolitem, -1)
            self.ccancel.set_sensitive(False)
            self.csave.set_sensitive(True)
          
        main_grid = self.builder.get_object("grid1")
        main_grid.set_row_homogeneous(True)
        main_grid.set_column_homogeneous(True)       
        # Insert Movebox
        if gv.show_moves == True:
            self.move_box = Gtk.ScrolledWindow()
            self.move_view = Gtk.TreeView()
            mlabel = Gtk.Label("Moves")
            self.move_box.add(self.move_view)             #omitted viewport
            main_grid.attach(self.move_box, 0, 15, 4, 5)  #Synthax: left, top, width, height
            self.move_box.set_policy(1,1)
            main_grid.attach(mlabel,0,14,4,1)
            #model
            self.movestore = Gtk.ListStore(GObject.TYPE_STRING)  #model for Treeview 
            self.move_view.set_model(self.movestore)
            self.move_view.set_tooltip_text(_("moves: double-click or click-enter to jump to"))
            self.comment_box = Gtk.ScrolledWindow()
            self.comment_view = Gtk.TextView()
            clabel = Gtk.Label("Comments")
            self.comment_box.add_with_viewport(self.comment_view)
            
            self.comment_view.set_wrap_mode(2)   #GTK_WRAP_WORD)
            g = self.comment_view.get_wrap_mode()
            self.comment_view.set_tooltip_text(_("comments")) 
            
            main_grid.attach(self.comment_box, 28, 1, 4, 5)        
            main_grid.attach(clabel,28,0,4,1)
            #self.move_view.set_editable(False)
            cell0 = Gtk.CellRendererText()
            # cell0.set_property("cell-background", Gdk.color_parse("#F8F8FF"))
            mvcolumn0 = Gtk.TreeViewColumn("#")
            self.move_view.append_column(mvcolumn0)
            mvcolumn0.pack_start(cell0, True)
            mvcolumn0.set_min_width(30)
            mvcolumn0.set_attributes(cell0, text=0)            
            self.comment_view.set_editable(False)
            self.comment_view.set_cursor_visible(False)
            #self.comment_view.connect("realized", self.realized)
            #self.move_view.activate_on_single_click(False)
            self.move_view.connect("cursor_changed", self.moves_clicked)
            #self.move_view.connect("key_press_event", self.moves_clicked)           #("row_activated", self.moves_clicked)   #"row_activated"

        # Create a 8x8 table for the main board
        self.boardgrid = Gtk.Grid.new()
        self.boardgrid.set_row_homogeneous(True)
        self.boardgrid.set_column_homogeneous(True)
        self.boardgrid.set_border_width(1)

        #
        # init board squares (add event box, set up drag and drop,
        # button clicks)
        # x, y = 0, 0 is the top left square of the board
        #
        for x in range(8):
            for y in range(8):
                self.init_board_square(x, y)

        eb2 = Gtk.EventBox()

        eb2.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse("black"))
        eb2.add(self.boardgrid)
        eb = Gtk.EventBox()
        eb.add(eb2)
        eb.show()

        aspect_frame = Gtk.AspectFrame(label=None, xalign=0.5, yalign=0.5,
                                       ratio=1.0, obey_child=False)
        aspect_frame.add(eb)

        eb2.set_border_width(20)
        self.grid_eb = eb2

        main_grid.attach(aspect_frame, 6, 0, 20, 20)
        #aspect_frame.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse("black"))

        eb.connect("draw", self.draw_border)
        self.border_eb = eb

        # status bar
        self.status_bar = self.builder.get_object("status_bar")

        # set status bar bg color
        # Use an event box and set its background colour.
        # This was needed on Fedora 19 to set the statusbar bg colour.
        # Otherwise it uses the window bg colour which is not
        # correct. This was not needed on F17.
        eb_2 = self.builder.get_object("eb_2")
        #eb_2.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse(bg_colour)) ##EDECEB
        self.status_bar = Gtk.Statusbar()
        main_vbox.pack_start(self.status_bar, False, False, 0)
        self.context_id = self.status_bar.get_context_id("jcchess statusbar")

        self.actiongroup.get_action("MoveNow").set_sensitive(False)

        self.window.show_all()
        self.side_to_move[BLACK].hide()
        self.gobutton.set_sensitive(False)
        self.stopbutton.set_sensitive(False)

        # mask = Gdk.WindowHints.BASE_SIZE|
        #            Gdk.WindowHints.MIN_SIZE|
        #            Gdk.WindowHints.MAX_SIZE|
        #            Gdk.WindowHints.RESIZE_INC|Gdk.WindowHints.ASPECT
        #mask = Gdk.WindowHints.MIN_SIZE | Gdk.WindowHints.MAX_SIZE
        mask = Gdk.WindowHints.MIN_SIZE
        geometry = Gdk.Geometry()
        # geometry.base_width = -1
        # geometry.base_height = -1
        # geometry.max_width = -1
        # geometry.max_height = -1
        geometry.min_width = 378
        geometry.min_height = 378
        # geometry.width_inc = -1
        # geometry.height_inc = -1
        # geometry.min_aspect = -1.0
        # geometry.max_aspect = -1.0
        #self.window.set_geometry_hints(self.window, geometry, mask)
        # self.window.set_geometry_hints(
        #   self.window, min_width=378, min_height=378, max_width=-1,
        #   max_height=-1, base_width=-1, base_height=-1, width_inc=-1,
        #   height_inc=-1, min_aspect=-1.0, max_aspect=-1.0)

        self.build_edit_popup()
    
       
    def moves_clicked(self, widget, data=None):
        self.moves_clicked_(0)
    
    def moves_clicked_(self, incr):
        model, triter = self.move_view.get_selection().get_selected()
        k = self.movestore.get_value(triter,0).find(".")      
        nmove = int(self.movestore.get_value(triter,0)[0:k])   
        self.move_list.set_move(nmove)                                            
        self.move_list.move_box_selection()
    
    def set_cedit(self, widget):
        self.cedit.set_sensitive(False)
        self.ccancel.set_sensitive(True)
        self.csave.set_sensitive(True)
        self.comment_view.set_editable(True)
        self.comment_view.cursor_visible = True
        
        
    
    def set_csave(self, widget):
        self.cedit.set_sensitive(True)
        self.ccancel.set_sensitive(False)
        self.csave.set_sensitive(False)
        self.comment_view.set_editable(False)
        self.comment_view.cursor_visible = False       
        model, triter = self.move_view.get_selection().get_selected()
        if triter != None:
            k = self.movestore.get_value(triter,0).find(".")      
            nmove = int(self.movestore.get_value(triter,0)[0:k])
            start = self.comment_view.get_buffer().get_start_iter()
            end = self.comment_view.get_buffer().get_end_iter()
            self.move_list.set_comment(nmove,self.comment_view.get_buffer().get_text(start, end, False))
            self.move_list.set_comment_ind(True)
        #print("comment ",nmove)
        #save
    
    def set_ccancel(self, widget):   
        self.cedit.set_sensitive(True)
        self.ccancel.set_sensitive(False)
        self.csave.set_sensitive(False) 
        self.comment_view.set_editable(False)
        self.comment_view.set_editable(False)
        self.comment_view.cursor_visible = False          

      
    def draw_board_square(self, widget, cr, x, y):
        gv.board.set_image_cairo(x, y, cr=cr, widget=widget)

    def init_board_square(self, x, y):
        event_box = self.eb[x][y]
        event_box.connect("draw", self.draw_board_square, x, y)
        # use 7-y since 0,0 is in bottom left corner for chessboard
        # but top left for grid 
        self.boardgrid.attach(event_box, x, 7-y, 1, 1)
        event_box.set_hexpand(True)
        event_box.set_vexpand(True)
        event_box.show()
        event_box.add_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        data = (x, y)
        event_box.connect("button_press_event", gv.jcchess.square_clicked, data)
        event_box.connect("drag_data_get", self.drag_and_drop.sendCallback)
        event_box.connect_after(
            "drag_begin", self.drag_and_drop.drag_begin, (x, y))
        event_box.connect_after("drag_end", self.drag_and_drop.drag_end)

        # set up square as a destination square to receive drag data
        # from a drag & drop action
        event_box.connect(
            "drag_data_received", self.drag_and_drop.receiveCallback, (x, y))

    def dnd_set_source_square(self, x, y):
        self.targets = [
            Gtk.TargetEntry.new(
                "text/plain", Gtk.TargetFlags.SAME_APP, TARGET_TYPE_TEXT)]
        self.eb[x][y].drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK, self.targets, Gdk.DragAction.COPY)

    def dnd_unset_source_square(self, x, y):
        self.eb[x][y].drag_source_unset()

    def dnd_set_dest_square(self, x, y):
        self.targets = [
            Gtk.TargetEntry.new(
                "text/plain", Gtk.TargetFlags.SAME_APP, TARGET_TYPE_TEXT)]
        self.eb[x][y].drag_dest_set(
            Gtk.DestDefaults.MOTION |
            #Gtk.DestDefaults.HIGHLIGHT |
            Gtk.DestDefaults.DROP,
            self.targets, Gdk.DragAction.COPY)

    def dnd_unset_dest_square(self, x, y):
        self.eb[x][y].drag_dest_unset()

    # set all squares so they cannot be used for drag and drop
    def unset_all_drag_and_drop_squares(self):
        # main board
        for x in range(8):
            for y in range(8):
                self.dnd_unset_source_square(x, y)
                self.dnd_unset_dest_square(x, y)

    #
    # Look at each board square and set it as a valid drag and drop
    # source square or target square if applicable
    #
    def apply_drag_and_drop_settings(self, player, stm):

        if gv.verbose:
            print("in apply_drag_and_drop_settings")

        # If drag and drop not enabled then exit
        # if not self.ta_action_group.get_action("enableDND").get_active():
        if not self.enable_dnd:
            self.unset_all_drag_and_drop_squares
            return

        for x in range(8):
            for y in range(8):
                # set default to unset (no square can be dragged or dropped
                self.dnd_unset_source_square(x, y)
                self.dnd_unset_dest_square(x, y)

                # if not human to move then no drag and drop allowed on any
                # square
                if player != "Human":
                    continue

                # player is human so allow a square to be dragged if it
                # contains a piece for his side

                # human piece - set as valid source square for dnd
                if gv.board.valid_source_square(x, y, stm):
                    self.dnd_set_source_square(x, y)
                    self.dnd_unset_dest_square(x, y)
                else:
                    # valid target square for dnd
                    self.dnd_unset_source_square(x, y)
                    self.dnd_set_dest_square(x, y)

    def get_event_box(self, x, y):
        return self.eb[x][y]

    # about box
    def about_box(self, widget):
        about = Gtk.AboutDialog(parent=self.window)
        #
        # set_program_name method is available in PyGTK 2.12 and above
        #
        try:
            about.set_program_name(NAME)
        except AttributeError:
            pass
        about.set_version(VERSION)
        about.set_copyright("Copyright \u00A9 2010-2016 John Cheetham")
        about.set_comments(
            "jcchess is a program to play chess.")
        about.set_authors(["John Cheetham"])
        about.set_website(
            "http://www.johncheetham.com/projects/jcchess/index.html")
        about.set_logo(
            GdkPixbuf.Pixbuf.new_from_file(
                os.path.join(gv.jcchess.prefix, "images/logo.png")))

        license = """jcchess is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

jcchess is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with jcchess.  If not, see <http://www.gnu.org/licenses/>."""

        about.set_license(license)
        about.run()
        about.destroy()

    def set_status_bar_msg(self, msg):
        if gv.jcchess.quitting:
            return
        self.context_id = self.status_bar.get_context_id("jcchess statusbar")
        self.status_bar.push(self.context_id, msg)
            
    # ask before promoting
    def promote_mode(self, w):
        if w.get_active():
            gv.jcchess.set_promotion_mode(True)
        else:
            gv.jcchess.set_promotion_mode(False)

    # disable some functionality if computer is thinking or in edit mode
    def disable_menu_items(self, mode=None):
        self.go_first.set_sensitive(False)
        self.go_back.set_sensitive(False)
        self.go_forward.set_sensitive(False)
        self.go_last.set_sensitive(False)

        self.actiongroup.get_action("NewGame").set_sensitive(False)
        self.actiongroup.get_action("NewHandicapGame").set_sensitive(False)
        self.actiongroup.get_action("LoadGame").set_sensitive(False)
        self.actiongroup.get_action("SaveGame").set_sensitive(False)

        if mode == "editmode":
            # edit menu
            self.actiongroup.get_action("CopyPosition").set_sensitive(False)
            self.actiongroup.get_action("PastePosition").set_sensitive(False)
            self.actiongroup.get_action("CopyGame").set_sensitive(False)
            self.actiongroup.get_action("PasteGame").set_sensitive(False)
            self.actiongroup.get_action("EditPosition").set_sensitive(True)

            # if edit mode disable all options menu
            self.actiongroup.get_action("Options").set_sensitive(False)
            self.gobutton.set_sensitive(False)
            self.stopbutton.set_sensitive(False)
        else:
            self.actiongroup.get_action("Edit").set_sensitive(False)
            # set items in options menu
            self.actiongroup.get_action("Undo").set_sensitive(False)
            self.actiongroup.get_action("Redo").set_sensitive(False)
            self.actiongroup.get_action("MoveNow").set_sensitive(True)
            self.actiongroup.get_action("SetBoardColours").set_sensitive(False)
            self.actiongroup.get_action("TimeControl").set_sensitive(False)
            self.actiongroup.get_action(
                "ConfigureEngine1").set_sensitive(False)
            self.actiongroup.get_action(
                "ConfigureEngine2").set_sensitive(False)
            self.actiongroup.get_action("Players").set_sensitive(False)
            self.actiongroup.get_action("Engines").set_sensitive(False)
            self.actiongroup.get_action(
                "CommonEngineSettings").set_sensitive(False)

            self.gobutton.set_sensitive(False)
            self.stopbutton.set_sensitive(True)

    # enable some functionality when computer is finished thinking
    def enable_menu_items(self, mode=None):
        self.go_first.set_sensitive(True)
        self.go_back.set_sensitive(True)
        self.go_forward.set_sensitive(True)
        self.go_last.set_sensitive(True)
        # self.ta_action_group.get_action("enableDND").set_sensitive(True)
        self.actiongroup.get_action("Undo").set_sensitive(True)
        self.actiongroup.get_action("Redo").set_sensitive(True)
        self.actiongroup.get_action("ConfigureEngine1").set_sensitive(True)
        self.actiongroup.get_action("ConfigureEngine2").set_sensitive(True)
        self.actiongroup.get_action("NewGame").set_sensitive(True)
        self.actiongroup.get_action("NewHandicapGame").set_sensitive(True)
        self.actiongroup.get_action("LoadGame").set_sensitive(True)
        self.actiongroup.get_action("SaveGame").set_sensitive(True)
        self.actiongroup.get_action("Engines").set_sensitive(True)
        self.actiongroup.get_action("Players").set_sensitive(True)
        self.actiongroup.get_action("TimeControl").set_sensitive(True)
        self.actiongroup.get_action("MoveNow").set_sensitive(False)
        self.actiongroup.get_action("CommonEngineSettings").set_sensitive(True)
        self.actiongroup.get_action("SetBoardColours").set_sensitive(True)

        self.actiongroup.get_action("Edit").set_sensitive(True)
        # edit menu
        self.actiongroup.get_action("CopyPosition").set_sensitive(True)
        self.actiongroup.get_action("PastePosition").set_sensitive(True)
        self.actiongroup.get_action("CopyGame").set_sensitive(True)
        self.actiongroup.get_action("PasteGame").set_sensitive(True)
        self.actiongroup.get_action("EditPosition").set_sensitive(True)

        self.gobutton.set_sensitive(True)
        self.stopbutton.set_sensitive(False)

        if mode == "editmode":
            # if edit mode disable all options menu
            self.actiongroup.get_action("Options").set_sensitive(True)

    def enable_go_button(self):
        self.gobutton.set_sensitive(True)

    def disable_go_button(self):
        self.gobutton.set_sensitive(False)

    def disable_stop_button(self):
        self.stopbutton.set_sensitive(False)

    def enable_stop_button(self):
        self.stopbutton.set_sensitive(True)

    def get_promotion_piece(self):
        return self.promotion_piece
        
    def promote_popup(self):
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.promotion_glade_file)
        self.builder.connect_signals(self)
        promotion_dialog = self.builder.get_object("promotion_dialog")
        promotion_dialog.set_transient_for(gv.gui.get_window())

        response_cancel = 1
        response_ok = 2

        # If user hasn't clicked on OK then exit now
        if promotion_dialog.run() != response_ok:
            promotion_dialog.destroy()
            # user cancelled so restore pieceset to what it was
            #gv.board.use_pieceset(self.orig_pieceset)
        else:
            promotion_dialog.destroy()
            
            
            
    # callback for when promotion piecset radiobutton changed
    def promotion_radio_button_changed(self, promotion_radio_button):
        #if not promotion_radio_button.get_active():
        #    return
        name = promotion_radio_button.get_label()
        if name == "Queen":
            self.promotion_piece = "q"
        elif name == "Rook":
            self.promotion_piece = "r"
        elif name == "Bishop":
            self.promotion_piece = "b"
        elif name == "Knight":
            self.promotion_piece = "n"
        else:
            print("invalid promotion in promotion_radio_button_changed in "
                  "gui.py:", name)
            self.promotion_piece = "q"

    def update_toolbar(self, player):
        self.engines_lblw.set_markup("<b>" + player[WHITE][:25] + " </b>")
        self.engines_lblb.set_markup("<b>" + player[BLACK][:25] + " </b>")

    #
    # Update the clocks on the display
    # These countdown the time while the player is thinking
    #
    def set_toolbar_time_control(self, txt, side_to_move):
        if side_to_move is None:
            return
        GLib.idle_add(self.tc_lbl[side_to_move].set_text, txt)

    def info_box(self, msg):
        dialog = Gtk.MessageDialog(
            self.window,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.INFO,
            Gtk.ButtonsType.OK,
            None)

        # markup = "<b>" + msg + "</b>"
        markup = msg
        dialog.set_markup(markup)
        dialog.set_title(_("Info"))

        # some secondary text
        markup = msg

        # dialog.format_secondary_markup(markup)

        dialog.show_all()
        dialog.run()
        dialog.destroy()

    def ok_cancel_box(self, msg):
        dialog = Gtk.MessageDialog(
            None,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
            Gtk.MessageType.INFO,
            Gtk.ButtonsType.OK_CANCEL,
            None)

        # markup = "<b>" + msg + "</b>"
        markup = msg
        dialog.set_markup(markup)
        dialog.set_title(_("Ok/Cancel"))

        # some secondary text
        markup = msg

        # dialog.format_secondary_markup(markup)

        dialog.show_all()
        resp = dialog.run()
        dialog.destroy()

        return resp

    def set_side_to_move(self, stm):
        self.side_to_move[stm].show()
        self.side_to_move[stm ^ 1].hide()

    # set the base window size at startup
    def set_window_size(self):
        if self.screen_width > 575 and self.screen_height > 724:
            w = 575
            h = 724
        elif self.screen_width > 415 and self.screen_height > 523:
            w = 415
            h = 523
        else:
            w = 322
            h = 406 
        #w = 300
        #h = 300
        #self.window.set_default_size(w, h)
        self.window.resize(w, h)

    def set_colours(self):
        if gv.verbose:
            print("in gui set_colours")
        bg_colour = '#888a85'
        self.get_window().modify_bg(
            Gtk.StateType.NORMAL, Gdk.color_parse(bg_colour))
        eb_2 = self.builder.get_object("eb_2")
        eb_2.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse(bg_colour))

        # square colours are set in board.py in set_image_cairo

        # border surrounds the board and contains the co-ordinates
        #self.border_eb.modify_bg(
        #    Gtk.StateType.NORMAL, Gdk.color_parse(border_colour))
        self.border_eb.queue_draw()

        grid_colour = '#000000'
        self.grid_eb.modify_bg(
            Gtk.StateType.NORMAL, Gdk.color_parse(grid_colour))

        gv.board.update()

    def build_edit_popup(self):
        self.edit_mode = False

        """
        popup_items = ["Separator",
                       _("Empty"),
                       _("Pawn"),
                       _("Knight"),
                       _("Bishop"),
                       _("Rook"),
                       _("Queen"),                
                       _("King"),
                       "Separator", 
                       _("Black to Move"),
                       _("White to Move"),
                       "Separator",
                       _("Clear Board"),
                       "Separator",
                       _("Cancel"),
                       _("End")]
        """                
        
        popup_items = ["Separator",
                       _("Empty"),
                       _("Pawn"),
                       _("Knight"),
                       _("Bishop"),
                       _("Rook"),
                       _("Queen"),                
                       _("King"),
                       "Separator", 
                       _("Black to Move"),
                       _("White to Move"),
                       "Separator",
                       _("Clear Board"),
                       "Separator",
                       _("End")]

        # set up menu for black
        self.bmenu = Gtk.Menu()
        menuitem = Gtk.MenuItem(_("Black"))
        menuitem.set_sensitive(False)
        self.bmenu.append(menuitem)
        # self.bmenu.append(Gtk.SeparatorMenuItem())
        for item_name in popup_items:
            if item_name == "Separator":
                self.bmenu.append(Gtk.SeparatorMenuItem())
                continue
            menuitem = Gtk.MenuItem(item_name)
            self.bmenu.append(menuitem)
            menuitem.connect("activate", self.edit_popup_callback, BLACK)

        # set up menu for white
        self.wmenu = Gtk.Menu()
        menuitem = Gtk.MenuItem(_("White"))
        menuitem.set_sensitive(False)
        self.wmenu.append(menuitem)
        # self.wmenu.append(Gtk.SeparatorMenuItem())
        for item_name in popup_items:
            if item_name == "Separator":
                self.wmenu.append(Gtk.SeparatorMenuItem())
                continue
            menuitem = Gtk.MenuItem(item_name)
            self.wmenu.append(menuitem)
            menuitem.connect("activate", self.edit_popup_callback, WHITE)

    # called when the user clicks on an item in the popup menu when editing
    # the board
    def edit_popup_callback(self, menuitem, colour):

        try:
            piece_name = menuitem.get_label()
        except AttributeError as ae:
            self.info_box(
                "Unable to edit board - you need a newer version of pygtk"
                "(version 2.16 or above)")
            return

        if piece_name == _("Clear Board"):
            gv.board.clear_board()
            return

        if piece_name == _("Black to Move"):
            gv.jcchess.set_side_to_move(BLACK)
            self.set_side_to_move(BLACK)   # update ind in gui
            gv.board.chessboard.turn=chess.BLACK
            return

        if piece_name == _("White to Move"):
            gv.jcchess.set_side_to_move(WHITE)
            self.set_side_to_move(WHITE)   # update ind in gui
            gv.board.chessboard.turn=chess.WHITE
            return

        if piece_name == _("Cancel"):
            self.edit_mode = False
            gv.jcchess.set_side_to_move(self.orig_stm)
            self.set_side_to_move(self.orig_stm)
            gv.board.update()            # restore board to its pre-edit state
            self.enable_menu_items(mode="editmode")
            return

        if piece_name == _("End"):
            self.end_edit()
            return

        # pieces contains the list of possible pieces in self.board_position
        # pieces = [
        #    " -", " p", " l", " n", " s", " g", " b", " r", " k", "+p", "+l",
        #    "+n", "+s", "+b", "+r", " P", " L", " N", " S", " G", " B", " R",
        #    " K", "+P", "+L", "+N", "+S", "+B", "+R"]

        #piece_dict = {
        #    _("Empty"): " -", _("Pawn"): "p", _("Lance"): " l", _("Knight"): " n",
        #    _("Silver"): " s", _("Gold"): " g", _("Bishop"): " b", _("Rook"): " r",
        #    _("King"): " k", _("+Pawn"): "+p", _("+Lance"): "+l", _("+Knight"): "+n",
        #    _("+Silver"): "+s", _("+Bishop"): "+b", _("+Rook"): "+r"} 
            
        if piece_name == _("Empty"):
            piece = None
        elif piece_name == _("Pawn"):
            piece = chess.PAWN
        elif piece_name == _("Knight"):
            piece = chess.KNIGHT
        elif piece_name == _("Bishop"):
            piece = chess.BISHOP
        elif piece_name == _("Rook"):
             piece = chess.ROOK
        elif piece_name == _("Queen"):
            piece = chess.QUEEN
        elif piece_name == _("King"):
            piece = chess.KING    
            

        if colour == WHITE:
            colour = chess.WHITE
        else:
            colour = chess.BLACK    
        # add piece to main board
        gv.board.set_piece_at_square(self.ed_x, self.ed_y, piece, colour)

    # save position and exit edit mode
    def end_edit(self):
        self.edit_mode = False
        fen = gv.board.get_fen()
        gv.load_save.init_game(fen)      # update board to reflect edit
        self.enable_menu_items(mode="editmode")

    def enable_edit_mode(self, action):
        # if already in edit mode then save edit position and exit edit mode
        if self.edit_mode:
            self.end_edit()
            return
        self.edit_mode = True
        self.orig_stm = gv.jcchess.get_side_to_move()
        self.set_status_bar_msg("Edit Mode")
        self.disable_menu_items(mode="editmode")

    def get_edit_mode(self):
        return self.edit_mode

    # called from jcchess.py when square clicked to popup the menu when
    # in edit board mode.
    def show_edit_popup(self, event, x, y):
        self.ed_x = x
        self.ed_y = y
        if event.button == 1:
            # left mouse button - display popup menu for black
            self.bmenu.popup(None, None, None, None, event.button, event.time)
            self.bmenu.show_all()
        elif event.button == 3:
            # right mouse button - display popup menu for white
            self.wmenu.popup(None, None, None, None, event.button, event.time)
            self.wmenu.show_all()
 
    # Fill in the colour or texture of the border that surrounds the board
    # and add co-ordinates
    def draw_border(self, widget, cr):
        a = widget.get_allocation()
        gv.set_board_colours.set_border_colour(cr, a)
        if not self.show_coords:
            return

        cr = widget.get_window().cairo_create()

        # cr.set_source_rgb(0.0, 0.0, 0.0)  # black
        col = '#000000'
        r = col[1:3]
        g = col[3:5]
        b = col[5:7]
        r_int = int(r, 16)
        g_int = int(g, 16)
        b_int = int(b, 16)
        cr.set_source_rgb(
            float(r_int) / 255, float(g_int) / 255, float(b_int) / 255)

        cr.select_font_face(
            "Monospace", cairo.FONT_SLANT_OBLIQUE, cairo.FONT_WEIGHT_NORMAL)

        tb_x = self.boardgrid.get_allocation().x
        tb_y = self.boardgrid.get_allocation().y
        tb_width = self.boardgrid.get_allocation().width
        tb_height = self.boardgrid.get_allocation().height

        sq_size = int(tb_width / 8)

        if sq_size > 75:
            font_size = 11
        elif sq_size > 56:
            font_size = 10
        elif sq_size > 40:
            font_size = 9
        elif sq_size > 30:
            font_size = 8
        else:
            font_size = 7

        cr.set_font_size(font_size)

        let = "abcdefgh"        
        xpos = int((widget.get_allocation().width - tb_width) / 2) + int(sq_size / 2)
        ypos1 = 14
        ypos2 = a.height - 7
        # show co-ordinate letters above and below the board
        for num in range(1, 9):
            # top of board
            cr.move_to(xpos, ypos1)
            cr.show_text(let[num - 1])
            # bottom of board
            cr.move_to(xpos, ypos2)
            cr.show_text(let[num - 1])
            xpos = xpos + sq_size
        
        # show co-ordinate numbers to left and right of the board
        # xpos = event.area.x + event.area.width - 14
        # xpos = widget.get_allocation().x = widget.get_allocation().width -14
        xpos1 = 7
        xpos2 = widget.get_allocation().width - 14
        # ypos = event.area.y + (
        #   event.area.height - tb_height) / 2 + sq_size / 2
        # ypos = widget.get_allocation().y + (
        #   widget.get_allocation().height - tb_height) / 2 + sq_size / 2
        ypos = int((widget.get_allocation().height - tb_height) / 2) + int(sq_size / 2)
        for num in range(1, 9):
            # left hand side of board
            cr.move_to(xpos1, ypos)
            cr.show_text(str(9 - num))
            # right hand side of board
            cr.move_to(xpos2, ypos)
            cr.show_text(str(9 - num))
            ypos = ypos + sq_size

    def preferences(self, action):
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file_preferences)
        self.builder.connect_signals(self)
        dialog = self.builder.get_object("preferences")
        dialog.set_transient_for(self.window)
        # show co-ords
        coords_checkbutton = self.builder.get_object("coords_checkbutton")
        coords_checkbutton.set_active(self.show_coords)

        # highlight moves
        highlight_moves_checkbutton = self.builder.get_object(
            "highlight_checkbutton")

        highlight_moves_checkbutton.set_active(self.highlight_moves)

        response = dialog.run()

        resp_cancel = 1
        resp_ok = 2
        if response == resp_ok:
            self.show_coords = coords_checkbutton.get_active()
            self.highlight_moves = highlight_moves_checkbutton.get_active()
            # rect = self.border_eb.get_allocation()
            # gv.board.update()
            self.border_eb.queue_draw()
        dialog.destroy()

    def set_show_coords(self, coords):
        self.show_coords = coords

    def set_highlight_moves(self, highlight_moves):
        self.highlight_moves = highlight_moves

    def get_highlight_moves(self):
        return self.highlight_moves

    def get_show_coords(self):
        return self.show_coords

    def get_window(self):
        return self.window



#
#   load_save.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#
from gi.repository import GLib
from gi.repository import Gtk
import os
from datetime import date
from . import gv
from . import move_list
from . import comments
from . import gamelist
import chess.pgn
from .constants import WHITE, BLACK,  VERSION,  NAME


class Load_Save:

    def __init__(self):
        self.move_list = move_list.get_ref()
        self.comments = comments.get_ref()
        self.gamelist = gamelist.get_ref()

    # Load game from a previously saved game
    def load_game(self, b):
        dialog = Gtk.FileChooserDialog(
            _("Load.."), gv.gui.get_window(), Gtk.FileChooserAction.OPEN,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
        dialog.set_default_response(Gtk.ResponseType.OK)
        #dialog.set_current_folder(os.path.expanduser("~"))
        dialog.set_current_folder(gv.lastdir)
        filter = Gtk.FileFilter()
        filter.set_name("pgn files")
        filter.add_pattern("*.pgn")
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)

        response = dialog.run()
        if response != Gtk.ResponseType.OK:
            dialog.destroy()
            return

        fname = dialog.get_filename()
        gv.lastdir = os.path.dirname(fname)
        if gv.verbose == True:       
            print ("opening: " + os.path.dirname(fname))
        gv.gui.window.set_title(NAME + " " + VERSION + "  " + os.path.basename(fname))
        dialog.destroy()

        if fname.endswith(".pgn"):
            self.get_header_from_file(fname)
            #entries from malformatted files would break the GUI of the program
            if gv.show_header == True:
                GLib.idle_add(gv.gui.header_lblsente.set_text, gv.sente[:50])
                #print(gv.sente)
                GLib.idle_add(gv.gui.header_lblgote.set_text, gv.gote[:50])
                #print(gv.gote)
                GLib.idle_add(gv.gui.header_lblevent.set_text, gv.event[:50])
                #print(gv.event)
                GLib.idle_add(gv.gui.header_lbldate.set_text, gv.gamedate[:50])
                #print(gv.gamedate)

            gamecnt = 0
            self.file_position = []
            self.gamelist.clear()
            
            # get headers and offsets of games
            f = open(fname)
            self.fname = fname
            for offset, headers in chess.pgn.scan_headers(f):
                gamecnt += 1
                self.gamelist.addgame(gamecnt, headers)
                self.file_position.append(offset)
            f.close()

            # No games in file
            if gamecnt == 0:
                gv.gui.info_box("No games in file")
                return

            # single game file - Load the game
            if gamecnt == 1:
                self.load_game_from_multigame_file(1)
                return

            # multi game file - Display the Game List so user can select a game
            self.gamelist.show_gamelist_window()

    # called from gamelist.py to load the game selected from the gamelist
    # of a multigame file
    def load_game_from_multigame_file(self, gameno):
        f = open(self.fname)
        f.seek(self.file_position[gameno - 1])
        game = chess.pgn.read_game(f)
        f.close()
        self.load_game_pgn(game)

    def load_game_pgn(self, game):
        stm = WHITE
        moveno = 0
        movelist = []
        redolist = []
        lastmove = ""
        startpos = "startpos"
        # if there is a FEN use that as board start position 
        # otherwise use standard start position
        try:
            fen = game.headers["FEN"]
        except KeyError:
            fen = "std"
        if gv.verbose:
            print("game=",game)
            print("fen=",fen)
        gv.board.init_board(fen)
        node = game
        while not node.is_end():
            next_node = node.variation(0)
            stm = stm ^ 1 
            move = node.board().san(next_node.move)
            if gv.verbose:
                print("move=", move)
                print("type=",type(move))
            move = gv.board.parse_san(move)
            if gv.verbose:
                print("move=", move)
                print("type=",type(move))
            gv.board.add_move(move)
            smove = str(move)                
            movelist.append(smove)
            lastmove = smove
            
            # add comment for this move if present
            if node.comment != "":
                self.comments.set_comment(moveno, node.comment)
                        
            moveno += 1
            if gv.verbose:
                gv.board.print_board()
            node = next_node
            
        gv.ucib.set_newgame()                    
        gv.uciw.set_newgame()
        gv.gui.set_status_bar_msg("game loaded")
        #self.gameover = False

        gv.jcchess.set_movelist(movelist)
        gv.jcchess.set_redolist(redolist)
        gv.jcchess.set_startpos(startpos)
        gv.jcchess.set_lastmove(lastmove)

        gv.board.update()

        # update move list in move list window
        self.move_list.update()
        stm = gv.jcchess.get_side_to_move()
        gv.jcchess.set_side_to_move(stm)
        gv.gui.set_side_to_move(stm)

        gv.tc.reset_clock()

        return 0

    #loads filename from 1st argument in commandline
    def load_game_parm(self,fname):
        try:
                fp = open(fname)
        except  :
                gv.gui.info_box("Error  loading game - file not found")
                return
        fp.close()
        gv.gui.window.set_title(NAME + " " + VERSION + "  " + os.path.basename(fname))
        
        if fname.endswith(".psn"):
            self.get_header_from_file(fname)
            if gv.show_header == True:
                GLib.idle_add(gv.gui.header_lblsente.set_text, gv.sente[:50])
                GLib.idle_add(gv.gui.header_lblgote.set_text, gv.gote[:50])
                GLib.idle_add(gv.gui.header_lblevent.set_text, gv.event[:50])
                GLib.idle_add(gv.gui.header_lbldate.set_text, gv.gamedate[:50])
            self.psn.load_game_psn(fname)
            return

    def get_header_from_file(self, fname):
        # sente,gote,event, date
        try:
                f = open(fname)
        except:
                gv.gui.info_box("Error  loading game-headers - file not found")
                
                return
        myList = []
        for line in f:
                myList.append(line)
        f.close()
        ff = False
        eventfound = False
        for line in myList:
                if line[0] == "[":
                        if line.find("Date") != -1:
                                gv.gamedate = line[6:-2]
                                #print("Date: ", gv.gamedate)
                                ff = True
                        if line.find("Event") !=-1:
                                gv.event = line[7:-2]
                                if gv.event=="" or gv.event=="##":
                                        gv.event="none"
                                #print(gv.event)
                                ff = True
                                eventfound = True
                               
                        if (line.find("Black")!=-1 or line.find("Sente")!=-1 ) and (line.find("SenteGrade")==-1) and (line.find("Black_grade")==-1):
                                        gv.sente =  line[8:-2] #"Sente:" +
                                        
                                        ff = True
                        if (line.find("White")!=-1 or line.find("Gote")!=-1) and (line.find("GoteGrade")==-1) and (line.find("White_grade")==-1):
                                gv.gote = line[7:-2]  #"Gote:" +
                                ff = True
        #if eventfound == False and myList[0][0] == "[":
         #       gv.event = myList[0][1:-1]              #educated guess
        #print("header from file:", gv.gamedate,  gv.gote,  gv.sente,  gv.event)
        if ff == True:
                # Header from file: mask in event removed
                if gv.event.find("##")!=-1:
                        gv.event = "no event"
                if gv.show_header == True:
                    GLib.idle_add( gv.gui.header_lbldate.set_text, gv.gamedate[:50])
                    GLib.idle_add( gv.gui.header_lblgote.set_text,  gv.gote[:50])
                    GLib.idle_add( gv.gui.header_lblsente.set_text, gv.sente[:50])
                    GLib.idle_add( gv.gui.header_lblevent.set_text, gv.event[:50])
        else:
                gv.event = "##"  # marks file without header
                
                
    # this routine is called from utils.py (when doing paste position)
    # and from gui.py (when ending an edit board session).
    def init_game(self, fen):
        gv.board.init_board(fen)
        startpos = fen
        #fenlst = fen.split()
        #if fenlst[1] == "b":
        #    if gv.verbose:
        #        print("setting stm to black")
        #    stm = BLACK
        #elif fenlst[1] == "w":
        #    stm = WHITE
        #else:
        #    stm = BLACK
        #engine.setplayer(stm)

        gv.ucib.set_newgame()
        gv.uciw.set_newgame()
        gv.gui.set_status_bar_msg(_("ready"))
        #self.gameover = False

        gv.jcchess.set_movelist([])
        gv.jcchess.set_redolist([])
        gv.jcchess.set_startpos(startpos)

        gv.board.update()

        # update move list in move list window
        self.move_list.update()

        stm = gv.jcchess.get_side_to_move()
        gv.jcchess.set_side_to_move(stm)
        gv.gui.set_side_to_move(stm)
        gv.jcchess.set_lastmove("")

        gv.tc.reset_clock()

    # called from utils.py as well as from this module
    def get_game(self):

        gamestr = ""
        startpos = gv.jcchess.get_startpos()
        # properties
        dat = str(date.today())
        dat = dat.replace("-", "/")

        zstr = '[Date "' + dat + '"]\n'
        gamestr += zstr

        zstr =  '[Sente: ' + gv.jcchess.get_player(BLACK).strip() + '"]\n'   
        gamestr += zstr

        zstr =  '[Gote: ' + gv.jcchess.get_player(WHITE).strip() + '"]\n' 
        gamestr += zstr

        # sfen
        if startpos == "startpos":
            zstr = '[SFEN "lnsgkgsnl/1r5b1/ppppppppp/9/9/9/PPPPPPPPP' \
                   '/1B5R1/LNSGKGSNL b - 1"]\n'
            gamestr += zstr
        else:
            zstr = '[SFEN "' + startpos + '"]\n'
            gamestr += zstr

        # add comment if present
        comment = self.comments.get_comment(0)
        if comment != "":
            gamestr = gamestr + "{\n" + comment + "\n}\n"

        # if the movelist is positioned part way through the game then
        # we must redo all moves so the full game will be saved
        redo_count = len(gv.jcchess.get_redolist())
        for i in range(0, redo_count):
            gv.jcchess.redo_move()

        # moves
        moveno = 1
        # ml = self.movelist
        mvstr = engine.getmovelist()

        # if we did any redo moves then undo them now to get things back
        # the way they were
        for i in range(0, redo_count):
            gv.jcchess.undo_move()

        if mvstr != "":
            mlst = mvstr.split(",")

            for m in mlst:

                (capture, ispromoted, move) = m.split(";")

                if move.find("*") == -1:
                    m1 = move[0:3]
                    m2 = move[3:]
                    move = m1 + capture + m2
                    if ispromoted == "+":
                        move = "+" + move
                zstr = str(moveno) + "." + move + "\n"
                gamestr += zstr
                # add comment for this move if present
                comment = self.comments.get_comment(moveno)
                if comment != "":
                    gamestr = gamestr + "{" + comment + "}\n"
                moveno += 1
        return (gamestr)

    # Save game to a file
    def save_game(self, b):

        dialog = Gtk.FileChooserDialog(
            _("Save.."), gv.gui.get_window(), Gtk.FileChooserAction.SAVE,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
        dialog.set_default_response(Gtk.ResponseType.OK)
        dialog.set_current_folder(gv.lastdir)

        filter = Gtk.FileFilter()
        filter.set_name("pgn files")
        filter.add_pattern("*.pgn")
        dialog.add_filter(filter)

        filter = Gtk.FileFilter()
        filter.set_name("All files")
        filter.add_pattern("*")
        dialog.add_filter(filter)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:

            startpos = gv.jcchess.get_startpos()

            #
            # filename must end with .pgn
            #
            filename = dialog.get_filename()
            gv.lastdir = os.path.dirname(filename)
            if gv.verbose == True:
                print("saving: " + gv.lastdir)
            gv.gui.window.set_title(NAME + " " + VERSION + "  " + os.path.basename(filename))
           
            if not filename.endswith('.pgn'):
                if dialog.get_filter().get_name() == "pgn files":
                    filename = filename + ".pgn" 

            # If file already exists then ask before overwriting
            if os.path.isfile(filename):
                resp = gv.gui.ok_cancel_box(
                    _("Warning - file already exists and will be replaced.\n"
                      "Press Cancel if you do not want to overwrite it."))
                if resp == Gtk.ResponseType.CANCEL:
                    dialog.destroy()
                    return
                    
            if filename.endswith(".pgn"):
                    
                game = gv.board.get_game()

                # add comments
                moveno = 0    
                node = game
                while not node.is_end():
                    next_node = node.variation(0)                                              
                    # add comment for this move if present
                    comment = self.comments.get_comment(moveno)
                    if comment != "":
                        node.comment = comment 
                    node = next_node
                    moveno += 1                
                   
                # write game to file                
                f = open(filename, "w",)
                print(game, file=f, end="\n\n")
                f.close()
                
            """
            if filename.endswith(".pgn"):
                # save in psn format
                gamestr = self.get_game()
                # we must get the last header read:
                gamelist = gamestr.splitlines(True) 
                #if gv.verbose == True:
                #print("Save psn:", gv.event,  gv.gamedate,  gv.sente,  gv.gote)
                if gv.event.find("##")==-1: 
                        #event not labelled with '##: 
                        #we must insert header from settings
                        #if gv.verbose == True:
                        #print( gv.event, "header rewritten")
                        gamelistnew = []
                        gamelistnew.append("[" + gv.event + "]\n")
                        gamelistnew.append("[Date: " + gv.gamedate+ "]\n")
                        gamelistnew.append("[Black:  " + gv.sente + "]\n")
                        gamelistnew.append("[White:  " + gv.gote + "]\n")
                        for item in gamelist[3:]:
                                gamelistnew.append(item)
                        gamestrnew =""
                        
                        for item in range(0, len(gamelistnew)):
                                gamestrnew = gamestrnew + str(gamelistnew[item])
                                
                        gamelistnew = ()
                        gamelist = ()
                        gamestr = gamestrnew
                
                f = open(filename, "w")
                f.write(gamestr)
                f.close()
            """
            
            gv.gui.set_status_bar_msg(_("game saved:  " + filename) + "  "+ VERSION)

        dialog.destroy()



#
#   utils.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import Gdk
import os
import sys
import pickle
import chess
import chess.pgn
from io import StringIO

from . import load_save
from . import gv
from .constants import VERSION


# Copy the board position to the clipboard in std FEN format
def copy_FEN_to_clipboard(action):
    fen = gv.board.get_fen()
    copy_text_to_clipboard(fen)


# paste a position from the clipboard
def paste_clipboard_to_FEN(action):
    fen = get_text_from_clipboard()
    if fen is None:
        gv.gui.info_box("Error: invalid fen")
        return
        
    # check the fen is valid    
    try:
        chessboard = chess.Board(fen)
    except ValueError as ve:
        print("Error - Pasted fen invalid: "+str(ve))
        print("Error - FEN data:",fen) 
        gv.gui.info_box("Error: invalid fen")
        return
    gv.load_save.init_game(fen)
    

def copy_game_to_clipboard(action):
    copy_text_to_clipboard(str(gv.board.get_game()))


def paste_game_from_clipboard(action):
    gamestr = get_text_from_clipboard()
    if gamestr is None:
        print("Error invalid game data")
        return
    pgn = StringIO(gamestr)
    game = chess.pgn.read_game(pgn)
    gv.load_save.load_game_pgn(game)

    
def copy_text_to_clipboard(text):
    # get the clipboard
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    # put the FEN data on the clipboard
    clipboard.set_text(text, -1)
    # make our data available to other applications
    clipboard.store()


def get_text_from_clipboard():
    # get the clipboard
    clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
    # read the text from the clipboard
    text = clipboard.wait_for_text()
    return text


def get_settings_from_file(filepath):
    s = None
    try:
        settings_file = os.path.join(filepath, "settings")
        f = open(settings_file, "rb")
        s = pickle.load(f)
        f.close()
    except EOFError as eofe:
        print("eof error:", eofe)
    except pickle.PickleError as pe:
        print("pickle error:", pe)
    except IOError as ioe:
        # Normally this error means it is the 1st run and the settings file
        # does not exist
        pass
    except Exception as exc:
        print("Cannot restore settings:", exc)
    #if gv.verbose:
    #    print "values read from settings file"
    #    print "colour_settings:",s.colour_settings
    return s


def get_prefix():
    # prefix to find package files/folders
    prefix = os.path.abspath(os.path.dirname(__file__))
    if gv.verbose:
        print("base directory (prefix) =", prefix)
    return prefix


def create_settings_dir():
    # set up jcchess directory under home directory
    home = os.path.expanduser("~")
    jcchesspath = os.path.join(home, ".jcchess")
    if not os.path.exists(jcchesspath):
        try:
            os.makedirs(jcchesspath)
        except OSError as exc:
            raise
    return jcchesspath


def get_verbose():
    verbose = False
    verbose_uci = False
    showmoves = False
    showheader = False
    for arg in sys.argv:
        if arg == "-v" or arg == "--verbose":
            verbose = True
        if arg == "-vuci":
            verbose_uci = True
        if arg == "-m":
            showmoves = True
        if arg == "-h":
            showheader = True
        if arg == "-mh" or arg == "-hm":
            showmoves = True
            showheader = True           
    return verbose, verbose_uci, showmoves, showheader



#
#   move_list.py - Display Move List Window
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
import os

from . import gv
#from . import gui
from . import comments

class Move_List:

    move_list_ref = None

    def __init__(self):
        glade_dir = gv.jcchess.get_glade_dir()
        self.glade_file = os.path.join(glade_dir, "move_list.glade")
        Move_List.move_list_ref = self
        self.comments = comments.get_ref()

        self.saved_move_list = []

        # create move list window
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file)
        self.builder.connect_signals(self)
        
        self.window = self.builder.get_object("move_list_window")
        self.vbox = self.builder.get_object("vbox1") #man
        self.treeview = Gtk.TreeView()  #man
        self.liststore = self.builder.get_object("liststore1")
        
        self.scrolled_window = Gtk.ScrolledWindow()
        #self.scrolled_window.set_size_request(150,300)
        self.treeview.set_model(self.liststore)
        self.scrolled_window.add(self.treeview)           
        self.vbox.add(self.scrolled_window)        
        self.comments_button = self.builder.get_object("comments_button")
        #self.builder.connect("button_press_event",self.comments_button_clicked_cb)
        
        cell0 = Gtk.CellRendererText()
        # cell0.set_property("cell-background", Gdk.color_parse("#F8F8FF"))
        tvcolumn0 = Gtk.TreeViewColumn("#")
        self.treeview.append_column(tvcolumn0)
        tvcolumn0.pack_start(cell0, True)
        tvcolumn0.set_min_width(50)
        tvcolumn0.set_attributes(cell0, text=0)

        cell1 = Gtk.CellRendererText()
        # cell1.set_property("cell-background", Gdk.color_parse("#F8F8FF"))
        tvcolumn1 = Gtk.TreeViewColumn(_("Move"))
        self.treeview.append_column(tvcolumn1)
        tvcolumn1.pack_start(cell1, True)
        tvcolumn1.set_min_width(100)
        tvcolumn1.set_attributes(cell1, text=1)

        cell2 = Gtk.CellRendererText()
        # cell1.set_property("cell-background", Gdk.color_parse("#F8F8FF"))
        tvcolumn2 = Gtk.TreeViewColumn(_("Cmt"))
        self.treeview.append_column(tvcolumn2)
        tvcolumn2.pack_start(cell2, True)
        tvcolumn2.set_min_width(20)
        tvcolumn2.set_attributes(cell2, text=2)
        
        
        self.tree_selection = self.treeview.get_selection()
        self.treeview.connect("button_press_event", self.treeview_button_press)
        self.treeview.connect("key_press_event", self.treeview_key_press)
        self.update()

    # user has closed the window
    # just hide it
    def delete_event(self, widget, event):
        self.window.hide()
        return True  # do not propagate to other handlers

    def show_movelist_window(self, b):
        # "present" will show the window if it is hidden
        # if not hidden it will raise it to the top
        self.window.show_all()
        self.window.present()
        return

    # update the move list
    # called when the number of moves in the list has changed
    def update(self):

        # update liststore
        self.liststore.clear()
        self.liststore.append(("0.", _("Start Pos"), " "))
        movelist = gv.jcchess.get_movelist()        
        moveno = 1
        if len(movelist) != 0:
            moveno = 1
            for move in movelist:
                m1 = move[0:2]
                m2 = move[2:]
                move = m1 + '-' + m2
                comment = self.comments.get_comment(moveno)
                if gv.show_moves == True:
                    gv.gui.comment_view.get_buffer().set_text("-")
                                    
                if comment != "":
                    cind = "..."
                else:
                    cind = " "
                if moveno % 2 == 1:
                    strmoveno = str(int((moveno + 1) / 2)) + "."
                else:
                    strmoveno = ""
                    
                #e = str(moveno) + ".", move, cind
                e = strmoveno, move, cind
                e1 = str(moveno) + "." + " " + move +" " + cind #+"\n"
                le = []
                le.append(e1)
                self.liststore.append(e)
                if gv.show_moves == True:
                    if moveno == 1:
                        gv.gui.move_view.get_model().clear()
                    gv.gui.move_view.get_model().append(le)
                                     
                     
                moveno += 1
 
        comment = self.comments.get_comment(moveno)
        if comment != "":
            if gv.show_moves == True:
                    gv.gui.comment_view.get_buffer().set_text(comment)
        GObject.idle_add(self.scroll_to_end)

    # sets the move at move_idx as the selected line
    # called from jcchess.py for undo/redo move
    def set_move(self, move_idx):
        path = (move_idx,)
        #self.tree_selection.select_path(path)  to be bypasses if cursor_changed as event in moves_clicked
        self.comments.set_moveno(move_idx)
        if gv.show_moves == True:
            if move_idx > 0:
                path = str(move_idx-1)
            #gv.gui.move_view.set_cursor(path, None,False) 
            if path[0]!=0 and path!='0':
                GObject.idle_add(gv.gui.move_view.scroll_to_cell,path,None, False, 0,0)  #arguments must be in list
            
        return

    def scroll_to_end(self):
        adj = self.scrolled_window.get_vadjustment()
        adj.set_value(adj.get_upper() - adj.get_page_size())
        return False

    def treeview_key_press(self, treeview, event):

        # to print list of key values do:
        #     print dir(Gtk.keysyms)

        # if up or down pressed then position the board at the move
        if event.keyval == Gdk.KEY_Up or event.keyval == Gdk.KEY_Down:
            self.treeview_button_press(None, None)

    # user clicked on the move list
    def treeview_button_press(self, treeview, event):
        if gv.jcchess.get_stopped():
            GObject.idle_add(self.process_tree_selection)
        else:
            GObject.idle_add(self.tree_selection.unselect_all)

    # set the board position at the move the user clicked on
    def move_box_selection(self):
        if gv.jcchess.get_stopped():
                       
            (treemodel, treeiter) = gv.gui.move_view.get_selection().get_selected()
            if treeiter is not None:
                k = gv.gui.movestore.get_value(treeiter,0).find(".")      
                nmove = int(gv.gui.movestore.get_value(treeiter,0)[0:k])
                self.comments.set_moveno(nmove)
                # now call a method in jcchess.py to position it at the move
                # clicked on
                gv.jcchess.goto_move(nmove)  
             
                path = str(nmove)
                self.treeview.set_cursor(path, None,False)  
                GObject.idle_add(self.treeview.scroll_to_cell,path,None, False, 0,0)  #arguments must be in list
                
        else:
            GObject.idle_add(gv.gui.move_view.unselect_all)            
    
    def process_tree_selection(self):
        (treemodel, treeiter) = self.tree_selection.get_selected()
        if treeiter is not None:
            
            # get the selected row number
            path = treemodel.get_path(treeiter)
            move_idx = path.get_indices()[0]
            
            #move_str = treemodel.get_value(treeiter, 0)
            #move_str = move_str[0: len(move_str) - 1]
            #move_idx = int(move_str)
            
            self.comments.set_moveno(move_idx)
            # now call a method in jcchess.py to position it at the move
            # clicked on
            gv.jcchess.goto_move(move_idx)

    def set_comment(self, index, text):
        self.comments.set_comment(index,text)
            

    def set_comment_ind(self, ind):
        if ind:
            cind = "..."
        else:
            cind = " "
        (treemodel, treeiter) = self.tree_selection.get_selected()
        if treeiter is not None:
            self.liststore.set_value(treeiter, 2, cind)

    def comments_button_clicked_cb(self, button):
        self.comments.show_comments_window()


def get_ref():
    if Move_List.move_list_ref is None:
        Move_List.move_list_ref = Move_List()
    return Move_List.move_list_ref



#
#   set_board_colours.py - Gui Dialog to Change the Bopard Colours
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
import cairo
import os

from . import gv


class Set_Board_Colours:

    BROWN, BLUE, GREEN = range(3)

    def __init__(self, prefix):
        if gv.verbose:
            print("in set_board_colours - init")
        self.colour_scheme = Set_Board_Colours.BROWN
        if gv.verbose:
            print("colour scheme set to default:", self.colour_scheme)
        # wood texture for border
        path = os.path.join(prefix, "images", "wood1.png")
        self.wood1 = cairo.ImageSurface.create_from_png(path)

    def set_colour_scheme(self, colour_scheme):
        if gv.verbose:
            print("in set_board_colours - set_colour_scheme")
            print("colour_scheme=", colour_scheme)
            
        if colour_scheme not in (0, 1, 2):
            self.colour_scheme = Set_Board_Colours.BROWN
            print("colour scheme not valid - set to default:",self.colour_scheme)
            return

        self.colour_scheme = colour_scheme

    def get_colour_scheme(self):
        return self.colour_scheme
        
    def get_square_colour(self):
        # darksquare / lightsquare 
        square_colours = (
          ( (205, 133,  63), (255, 222, 173) ),  # brown (default style)
          ( (131, 165, 210), (255, 255, 250) ),  # blue (droidfish style)
          ( (112, 160, 104), (200, 192, 96)  )   # green (xboard style)
        )
        return square_colours[self.colour_scheme]

    def set_border_colour(self, cr, a): 
           image = self.wood1
           cr.set_source_surface(image, 0, 0)
           cairo.Pattern.set_extend(cr.get_source(), cairo.EXTEND_REPEAT)
           cr.rectangle(0, 0 , a.width, a.height)
           cr.fill()

    def show_dialog(self, gtkaction):
        dialog = Gtk.Dialog(
            _("Select Colour Scheme"), gv.gui.get_window(), 0,
            (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
             Gtk.STOCK_OK, Gtk.ResponseType.OK))
        rb1 = Gtk.RadioButton.new_with_label(None, "Brown")
        dialog.vbox.pack_start(rb1, False, True, 5)
        rb2 = Gtk.RadioButton.new_with_label_from_widget(rb1, "Blue")
        dialog.vbox.pack_start(rb2, False, True, 5)
        rb3 = Gtk.RadioButton.new_with_label_from_widget(rb1, "Green")
        dialog.vbox.pack_start(rb3, False, True, 5)
        dialog.show_all()
        dialog.set_default_response(Gtk.ResponseType.OK)
        if self.colour_scheme == 0:
            rb1.set_active(True)
        elif self.colour_scheme == 1:
            rb2.set_active(True)
        elif self.colour_scheme == 2:
            rb3.set_active(True)
        else:
            rb1.set_active()
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            if rb1.get_active():
                self.colour_scheme = 0
            elif rb2.get_active():
                self.colour_scheme = 1
            elif rb3.get_active():
                self.colour_scheme = 2
        dialog.destroy()
        return

    def apply_colour_settings(self):
        if gv.verbose:
            print("set_board_colours - apply_colour_settings")
        gv.gui.set_colours()



#
#   comments.py - View/Edit Comments
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

from gi.repository import Gtk
import os

import jcchess.move_list
from . import gv


class Comments:

    comments_ref = None

    def __init__(self):

        glade_dir = gv.jcchess.get_glade_dir()
        self.glade_file = os.path.join(glade_dir, "comments.glade")
        self.move_list = jcchess.move_list.get_ref()

        # create comments window
        self.builder = Gtk.Builder()
        self.builder.set_translation_domain(gv.domain)
        self.builder.add_from_file(self.glade_file)
        self.builder.connect_signals(self)

        self.window = self.builder.get_object("comments_window")
        self.window.hide()

        self.moveno = 0  # the moveno the comment applies to

        tv = self.builder.get_object("comments_textview")
        tv.set_editable(True)
        tv.set_wrap_mode(Gtk.WrapMode.WORD)
        self.tb = tv.get_buffer()
        self.tb.connect("changed", self.text_changed)
        self.comment_list = []

    # user has closed the window
    # just hide it
    def delete_event(self, widget, event):
        self.window.hide()
        return True  # do not propagate to other handlers

    # called from comments_button_clicked_cb in move_list.py
    def show_comments_window(self):
        # "present" will show the window if it is hidden
        # if not hidden it will raise it to the top
        self.window.present()
        return

    # Called from move_list.py when selected move changes
    # change the comment window to show the comment for the
    # selected move
    def set_moveno(self, moveno):
        max = moveno
        if max < self.moveno:
            max = self.moveno
        # extend comment list if it is shorter than movelist
        while len(self.comment_list) <= max:
            self.comment_list.append("")

        # get current text from comments window
        start_iter = self.tb.get_start_iter()
        end_iter = self.tb.get_end_iter()
        text = self.tb.get_text(start_iter, end_iter, False)
        # save it
        # self.comment_list[self.moveno] = text

        self.moveno = moveno

        # set the comment in the window to that of the newly selected
        # move
        self.tb.set_text(self.comment_list[moveno])
        if gv.show_moves == True:
            start, end =gv.gui.comment_view.get_buffer().get_bounds()
            gv.gui.comment_view.get_buffer().delete(start,end)
            gv.gui.comment_view.get_buffer().insert(start,text)

        # show the moveno the comment relates to in the window title
        self.window.set_title(_("Comment for Move ") + gv.jcchess.convert_move(moveno))

        # self.moveno = moveno
    def get_comment_text(self,moveno):
        return self.comment_list[moveno]
        
    def automatic_comment(self,text,moveno):
        max = moveno
        if max < self.moveno:
            max = self.moveno
        # extend comment list if it is shorter than movelist
        while len(self.comment_list) <= max:
            self.comment_list.append("")        
        # extend comment list if it is shorter than movelist
        while len(self.comment_list) <= self.moveno:
            self.comment_list.append("")

        # save it
        # print "saving to move;",self.moveno
        # print "text=",text
        text1 = self.comment_list[self.moveno]
        self.comment_list[self.moveno] = text1 + text  #append comments
        if text != "":
            self.move_list.set_comment_ind(True)
        else:
            self.move_list.set_comment_ind(False)        
    
    def text_changed(self, textbuffer):
        # get text
        start_iter = textbuffer.get_start_iter()
        end_iter = textbuffer.get_end_iter()
        text = textbuffer.get_text(start_iter, end_iter, False)

        # extend comment list if it is shorter than movelist
        while len(self.comment_list) <= self.moveno:
            self.comment_list.append("")

        # save it
        # print "saving to move;",self.moveno
        # print "text=",text
        self.comment_list[self.moveno] = text
        if text != "":
            self.move_list.set_comment_ind(True)
        else:
            self.move_list.set_comment_ind(False)

    # Clear the comment window when clear button is clicked
    def clear_button_clicked_cb(self, button):
        text = ""
        # extend comment list if it is shorter than movelist
        while len(self.comment_list) <= self.moveno:
            self.comment_list.append("")
        self.comment_list[self.moveno] = text
        self.tb.set_text(text)
        if gv.show_moves == True:
            start, end =gv.gui.comment_view.get_buffer().get_bounds()
            gv.gui.comment_view.get_buffer().delete(start,end)
            gv.gui.comment_view.get_buffer().insert(start,"-")        
    # clear all comments.
    # Called from load_save.py when a new game is loaded
    def clear_comments(self):
        self.moveno = 0
        self.comment_list = [""]
        self.tb.set_text("")
        self.window.set_title(_("Comment for Move ") + "0")
        if gv.show_moves == True:
            start, end =gv.gui.comment_view.get_buffer().get_bounds()
            gv.gui.comment_view.get_buffer().delete(start,end)
            gv.gui.comment_view.get_buffer().insert(start,"-")           
    # called by load_save.py when loading file in PSN format
    def set_comment(self, moveno, comment):
        while len(self.comment_list) <= moveno:
            self.comment_list.append("")
        self.comment_list[moveno] = comment

    # called by load_save.py when saving file in PSN format
    def get_comment(self, moveno):
        while len(self.comment_list) <= moveno:
            self.comment_list.append("")
        return self.comment_list[moveno]

    # called by load_save.py when saving file to check if there are
    # comments on it when saving in gshog format
    def has_comments(self):
        for comment in self.comment_list:
            if len(comment) > 0:
                return True
        return False


def get_ref():
    if Comments.comments_ref is None:
        Comments.comments_ref = Comments()
    return Comments.comments_ref



#
#   board.py
#
#   This file is part of jcchess
#
#   jcchess is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   jcchess is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with jcchess.  If not, see <http://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Rsvg', '2.0')
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import Rsvg
import cairo
import socket
from datetime import datetime

import chess
import chess.pgn
from . import gv
from .constants import WHITE, BLACK

SCALE = 0.9      # scale the pieces so they occupy 90% of the board square
                 # (Also used in drag_and_drop.py)
LINEWIDTH = 2    # width of lines on the board


class Board:

    def __init__(self):
        self.init_board()
        self.dnd = None        

    def init_board(self, fen="std"):
        if fen == "std":
            self.chessboard = chess.Board() # init board
        else:
            self.chessboard = chess.Board(fen)    
        
    def build_board(self):
        GObject.idle_add(self.update)

    # convert jcchess co-ordinates for square into
    # standard notation (e.g.  (1, 0) -> b1)
    def get_square_posn(self, x, y):
        lets = "abcdefghi"
        sq = lets[x] + str(y+1) 
        return sq

    # convert standard notation for square into
    # jcchess co-ordinates (e.g.  b1 -> (1, 0) )
    def get_gs_square_posn(self, sq):        
        lets = "abcdefgh"
        x = lets.index(sq[0:1])
        y = int(sq[1:2]) - 1
        return x, y
 
    #
    # test if user has clicked on a valid source square
    # i.e. one that contains a black piece if side to move is black
    #      otherwise white piece
    def valid_source_square(self, x, y, stm):
        piece = self.get_piece(x, y)
        pieces = [
            [
               "r", "n", "b", "q", "k", "p"
            ],
            [
                "R", "N", "B", "Q", "K", "P"
            ]
        ]

        try:
            idx = pieces[stm].index(piece)
        except ValueError as ve:
            return False

        return True

    def use_pieceset(self, pieceset):
        gv.pieces.set_pieceset(pieceset)
        self.update()

    def update(self, squares_to_hilite=None):
        self.squares_to_hilite = squares_to_hilite
        #
        # loop through the board squares and set the pieces
        # x, y = 0, 0 is the top left square of the board
        #
        for x in range(8):
            for y in range(8):
                gv.gui.get_event_box(x, y).queue_draw()

    #
    # get an Rsvg.Handle of the piece at the given square
    # used by drag_and_drop.py to get the drag and drop icon
    # and by this module to draw pieces on the board
    #
    def get_piece_handle(self, x, y):
        piece = self.get_piece(x, y)
        handle = gv.pieces.gethandle(piece)
        return handle

    #
    # called from drag_and_drop.py to clear the source square when a
    # drag of a piece has started
    #
    def set_square_as_unoccupied(self, x, y):
        self.dnd = (x, y)
        # redraw square
        GLib.idle_add(gv.gui.get_event_box(x, y).queue_draw)

    # called from gui.py when editing the board position to set the piece
    # on a square.
    def set_piece_at_square(self, x, y, piece, colour):
        self.chessboard.set_piece_at(chess.square(x, y), chess.Piece(piece, colour))       
        GLib.idle_add(gv.gui.get_event_box(x, y).queue_draw)
        
    # called from gui.py to remove piece during promotion dialog
    def remove_piece_at_square(self, x, y):
        piece=self.chessboard.remove_piece_at(chess.square(x, y))       
        GLib.idle_add(gv.gui.get_event_box(x, y).queue_draw)
        return piece

    # called when user does a "clear board" in board edit
    def clear_board(self):
        self.chessboard.clear()
        self.update()

    def set_image_cairo(self, x, y, cr=None, widget=None):
        piece = self.get_piece(x, y)

        if cr is None:
            w = gv.gui.get_event_box(x, y).get_window()
            cr = w.cairo_create()

        if widget is not None:
            a = widget.get_allocation()
        else:
            a = gv.gui.get_event_box(x, y).get_allocation()

        # if user has set hilite moves on then check if this square is
        # in last move and if so hilight it
        hilite = False
        if gv.gui.get_highlight_moves():
            if self.squares_to_hilite is not None:
                if chess.square(x, y) in self.squares_to_hilite:
                   hilite = True
            else:
                # if squares to hilite not set then hilite last move       
                try:
                    lastmove = self.chessboard.peek()
                except IndexError:
                    lastmove = ""
                if lastmove != "":
                    sqnum = chess.square(x, y)
                    if sqnum in (lastmove.from_square, lastmove.to_square):
                        hilite = True

        # clear square to square colour
        darksquare, lightsquare = gv.set_board_colours.get_square_colour()
        if (x+y) % 2 == 0:
            r, g, b = darksquare
        else:
            r, g, b = lightsquare
        r, g, b = r/255, g/255, b/255                    

        # if square is highlighted put a red rectangle around it
        if hilite:
            cr.set_source_rgb(1, 0, 0)
            cr.rectangle(1, 1 , a.width, a.height)
            cr.fill()
            cr.set_source_rgb(r, g, b)
            cr.rectangle(3, 3, a.width-6, a.height-6)
            cr.fill()
        else:
            cr.set_source_rgb(r, g, b)
            cr.rectangle(1, 1 , a.width-LINEWIDTH, a.height-LINEWIDTH)
            cr.fill()
        if piece == "None":
            return
        # set offset so piece is centered in the square
        cr.translate(a.width*(1.0-SCALE)/2.0, a.height*(1.0-SCALE)/2.0)

        # scale piece so it is smaller than the square
        if self.dnd is not None and self.dnd == (x, y):
            self.dnd = None
            return
        else:    
            svghandle = self.get_piece_handle(x, y) 
        dim = svghandle.get_dimensions()
        sfw = (a.width * 1.0 / dim.width) * SCALE
        sfh = (a.height * 1.0 / dim.height) * SCALE
        cr.scale(sfw, sfh)
        svghandle.render_cairo(cr)

    def get_piece(self, x, y):
        piece = self.chessboard.piece_at(chess.square(x, y))
        piece = str(piece)   
        return piece

    def get_legal_moves(self):
        return self.chessboard.legal_moves
        
    def add_move(self, cmove):    
        self.chessboard.push(cmove)
        
    def remove_move(self):
        return self.chessboard.pop()
        
    def print_board(self):
        print("board fen:",repr(self.chessboard))
        print("board:\n",self.chessboard)
        
    def is_gameover(self):
        if self.chessboard.is_game_over():
            return True
        else:
            return False
                
    #def is_checkmate(self):
    #    if self.chessboard.is_checkmate():
    #        return True
    #    else:
    #        return False
            
    def parse_san(self, move):
        return self.chessboard.parse_san(move)
        
    def get_game(self):
        
        # if the movelist is positioned part way through the game then
        # we must redo all moves to get the full game
        redo_count = len(gv.jcchess.get_redolist())
        for i in range(0, redo_count):
            gv.jcchess.redo_move()
        
        game = chess.pgn.Game.from_board(self.chessboard)
        
        # if we did any redo moves then undo them now to get things back
        # the way they were
        for i in range(0, redo_count):
            gv.jcchess.undo_move()

        game.headers["Event"] = "Computer Chess Game"
        game.headers["Site"] = socket.gethostname()
        game.headers["Date"] = datetime.strftime(datetime.now(), '%Y.%m.%d')  
        game.headers["Round"] = "-"
        game.headers["White"] = gv.jcchess.get_player(WHITE)
        game.headers["Black"] = gv.jcchess.get_player(BLACK)
        return game
        
    def get_fen(self):
        return self.chessboard.fen()
                



#!/usr/bin/python

#
# start jcchess from within the source folder
#

import sys

assert sys.version_info >= (3,0)

import jcchess.gv

#sys.path.append("jcchess")
jcchess.gv.installed = False

import jcchess.jcchess
jcchess.jcchess.run()
